import numpy as np
import json
import argparse
from torch.utils.data import Dataset
import dgl
import dgl.function as fn
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from itertools import combinations
import time
from time import localtime, strftime
import os
import pickle
from scipy import sparse
from sklearn.cluster import KMeans
from sklearn import metrics
import sys
import pandas as pd
import en_core_web_lg
from datetime import datetime
import networkx as nx
from dgl.dataloading import MultiLayerNeighborSampler, NodeDataLoader
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from dataset.dataloader import DatasetLoader

"""
    KPGNN Model
    Paper: Knowledge-Preserving Incremental Social Event Detection via Heterogeneous GNNs
    Author: Yuwei Cao et al.
    github: https://github.com/RingBDStack/KPGNN
"""

class args_define:
    def __init__(self, **kwargs):
        # Hyper-parameters
        self.n_epochs = kwargs.get('n_epochs', 15)
        self.n_infer_epochs = kwargs.get('n_infer_epochs', 0)
        self.window_size = kwargs.get('window_size', 3)
        self.patience = kwargs.get('patience', 5)
        self.margin = kwargs.get('margin', 3.0)
        self.lr = kwargs.get('lr', 1e-3)
        self.batch_size = kwargs.get('batch_size', 200)
        self.n_neighbors = kwargs.get('n_neighbors', 800)
        self.hidden_dim = kwargs.get('hidden_dim', 8)
        self.out_dim = kwargs.get('out_dim', 32)
        self.num_heads = kwargs.get('num_heads', 4)
        self.use_residual = kwargs.get('use_residual', True)
        self.validation_percent = kwargs.get('validation_percent', 0.2)
        self.use_hardest_neg = kwargs.get('use_hardest_neg', False)
        self.use_dgi = kwargs.get('use_dgi', False)
        self.remove_obsolete = kwargs.get('remove_obsolete', 2)
        self.is_incremental = kwargs.get('is_incremental', False)

        # Other arguments
        self.use_cuda = kwargs.get('use_cuda', False)
        self.data_path = kwargs.get('data_path', '../model/model_saved/kpgnn/kpgnn_incremental_test')
        self.mask_path = kwargs.get('mask_path', None)
        self.resume_path = kwargs.get('resume_path', None)
        self.resume_point = kwargs.get('resume_point', 0)
        self.resume_current = kwargs.get('resume_current', True)
        self.log_interval = kwargs.get('log_interval', 10)

        # Store all arguments in a single attribute
        self.args = argparse.Namespace(**{
            'n_epochs': self.n_epochs,
            'n_infer_epochs': self.n_infer_epochs,
            'window_size': self.window_size,
            'patience': self.patience,
            'margin': self.margin,
            'lr': self.lr,
            'batch_size': self.batch_size,
            'n_neighbors': self.n_neighbors,
            'hidden_dim': self.hidden_dim,
            'out_dim': self.out_dim,
            'num_heads': self.num_heads,
            'use_residual': self.use_residual,
            'validation_percent': self.validation_percent,
            'use_hardest_neg': self.use_hardest_neg,
            'use_dgi': self.use_dgi,
            'remove_obsolete': self.remove_obsolete,
            'is_incremental': self.is_incremental,
            'use_cuda': self.use_cuda,
            'data_path': self.data_path,
            'mask_path': self.mask_path,
            'resume_path': self.resume_path,
            'resume_point': self.resume_point,
            'resume_current': self.resume_current,
            'log_interval': self.log_interval,
        })

'''
class args_define():
    parser = argparse.ArgumentParser()
    # Hyper-parameters
    parser.add_argument('--n_epochs', default=15, type=int,
                        help="Number of initial-training/maintenance-training epochs.")
    parser.add_argument('--n_infer_epochs', default=0, type=int,
                        help="Number of inference epochs.")
    parser.add_argument('--window_size', default=3, type=int,
                        help="Maintain the model after predicting window_size blocks.")
    parser.add_argument('--patience', default=5, type=int,
                        help="Early stop if performance did not improve in the last patience epochs.")
    parser.add_argument('--margin', default=3., type=float,
                        help="Margin for computing triplet losses")
    parser.add_argument('--lr', default=1e-3, type=float,
                        help="Learning rate")
    parser.add_argument('--batch_size', default=200, type=int,
                        help="Batch size (number of nodes sampled to compute triplet loss in each batch)")
    parser.add_argument('--n_neighbors', default=800, type=int,
                        help="Number of neighbors sampled for each node.")
    parser.add_argument('--hidden_dim', default=8, type=int,
                        help="Hidden dimension")
    parser.add_argument('--out_dim', default=32, type=int,
                        help="Output dimension of tweet representations")
    parser.add_argument('--num_heads', default=4, type=int,
                        help="Number of heads in each GAT layer")
    parser.add_argument('--use_residual', dest='use_residual', default=True,
                        action='store_false',
                        help="If true, add residual(skip) connections")
    parser.add_argument('--validation_percent', default=0.2, type=float,
                        help="Percentage of validation nodes(tweets)")
    parser.add_argument('--use_hardest_neg', dest='use_hardest_neg', default=False,
                        action='store_true',
                        help="If true, use hardest negative messages to form triplets. Otherwise use random ones")
    parser.add_argument('--use_dgi', dest='use_dgi', default=False,
                        action='store_true',
                        help="If true, add a DGI term to the loss. Otherwise use triplet loss only")
    parser.add_argument('--remove_obsolete', default=2, type=int,
                        help="If 0, adopt the All Message Strategy: keep inseting new message blocks and never remove;\n" +
                             "if 1, adopt the Relevant Message Strategy: remove obsolete training nodes (that are not connected to the new messages arrived during the last window) during maintenance;\n" +
                             "if 2, adopt the Latest Message Strategy: during each prediction, use only the new data to construct message graph, during each maintenance, use only the last message block arrived during the last window for continue training.\n" +
                             "See Section 4.4 of the paper for detailed explanations of these strategies. Note the message graphs for 0/1 and 2 need to be constructed differently, see the head comment of custom_message_graph.py")
    parser.add_argument('--is_incremental', action='store_true', default=False,
                        help="static or incremental")
    # Other arguments
    parser.add_argument('--use_cuda', dest='use_cuda', default=False,
                        action='store_true',
                        help="Use cuda")
    parser.add_argument('--data_path', default='../model/model_saved/kpgnn/kpgnn_incremental_test',
                        # default='./incremental_0808/',
                        type=str, help="Path of features, labels and edges")
    # format: './incremental_0808/incremental_graphs_0808/embeddings_XXXX'
    parser.add_argument('--mask_path', default=None,
                        type=str, help="File path that contains the training, validation and test masks")
    # format: './incremental_0808/incremental_graphs_0808/embeddings_XXXX'
    parser.add_argument('--resume_path', default=None,
                        type=str,
                        help="File path that contains the partially performed experiment that needs to be resume.")
    parser.add_argument('--resume_point', default=0, type=int,
                        help="The block model to be loaded.")
    parser.add_argument('--resume_current', dest='resume_current', default=True,
                        action='store_false',
                        help="If true, continue to train the resumed model of the current block(to resume a partally trained initial/mantenance block);\
                            If false, start the next(infer/predict) block from scratch;")
    parser.add_argument('--log_interval', default=10, type=int,
                        help="Log interval")

    args = parser.parse_args()
'''

class KPGNN():
    def __init__(self, args, dataset):
        self.args = args
        self.dataset = dataset
        self.model = None
        self.loss_fn = None
        self.loss_fn_dgi = None
        self.metrics = None
        self.train_indices = None
        self.indices_to_remove = None
        self.embedding_save_path = None
        self.data_split = None

    def preprocess(self):
        preprocessor = Preprocessor(self.args, self.dataset)
        # preprocessor.generate_initial_features(self.dataset)
        preprocessor.custom_message_graph(self.dataset)

    def fit(self):
        use_cuda = self.args.use_cuda and torch.cuda.is_available()
        print("Using CUDA:", use_cuda)
        os.makedirs(self.args.data_path, exist_ok=True)

        # make dirs and save args
        if self.args.resume_path is None:  # build a new dir if training from scratch
            self.embedding_save_path = self.args.data_path + '/embeddings_' + strftime("%m%d%H%M%S", localtime())
            os.mkdir(self.embedding_save_path)

        # resume training using original dir
        else:
            self.embedding_save_path = self.args.resume_path
        print("embedding_save_path: ", self.embedding_save_path)

        with open(self.embedding_save_path + '/args.txt', 'w') as f:
            json.dump(self.args.__dict__, f, indent=2)

        # Load data splits
        self.data_split = np.load(self.args.data_path + '/data_split.npy')

        # Loss
        if self.args.use_hardest_neg:
            self.loss_fn = OnlineTripletLoss(self.args.margin, HardestNegativeTripletSelector(self.args.margin))
        else:
            self.loss_fn = OnlineTripletLoss(self.args.margin, RandomNegativeTripletSelector(self.args.margin))
        if self.args.use_dgi:
            self.loss_fn_dgi = torch.nn.BCEWithLogitsLoss()

        self.metrics = [AverageNonzeroTripletsMetric()]

        train_i = 0

        if ((self.args.resume_path is not None) and (self.args.resume_point == 0) and (
                self.args.resume_current)) or self.args.resume_path is None:
            if not self.args.use_dgi:
                self.train_indices, self.indices_to_remove, self.model = KPGNN_model.initial_maintain(train_i, 0,
                                                                                                      self.data_split,
                                                                                                      self.metrics,
                                                                                                      self.embedding_save_path,
                                                                                                      self.loss_fn,
                                                                                                      self.model)
            else:
                self.train_indices, self.indices_to_remove, self.model = KPGNN_model.initial_maintain(train_i, 0,
                                                                                                      self.data_split,
                                                                                                      self.metrics,
                                                                                                      self.embedding_save_path,
                                                                                                      self.loss_fn,
                                                                                                      None,
                                                                                                      self.loss_fn_dgi)

    def detection(self):

        train_i = 0
        if self.args.is_incremental:

            # Initialize the model, train_indices and indices_to_remove to avoid errors
            if self.args.resume_path is not None:
                self.model = None
                self.train_indices = None
                self.indices_to_remove = []

            # iterate through all blocks
            for i in range(1, self.data_split.shape[0]):
                # Inference (prediction)
                # Resume model from the previous, i.e., (i-1)th block or continue the new experiment. Otherwise (to resume from other blocks) skip this step.
                if ((self.args.resume_path is not None) and (self.args.resume_point == i - 1) and (
                        not self.args.resume_current)) or self.args.resume_path is None:
                    if not self.args.use_dgi:
                        self.model = KPGNN_model.infer(train_i, i, self.data_split, self.metrics,
                                                       self.embedding_save_path, self.loss_fn, self.train_indices,
                                                       self.model, None,
                                                       self.indices_to_remove)
                    else:
                        self.model = KPGNN_model.infer(train_i, i, self.data_split, self.metrics,
                                                       self.embedding_save_path, self.loss_fn, self.train_indices,
                                                       self.model,
                                                       self.loss_fn_dgi, self.indices_to_remove)
                # Maintain
                # Resume model from the current, i.e., ith block or continue the new experiment. Otherwise (to resume from other blocks) skip this step.
                if ((self.args.resume_path is not None) and (self.args.resume_point == i) and (
                        self.args.resume_current)) or self.args.resume_path is None:
                    if i % self.args.window_size == 0:
                        train_i = i
                        if not self.args.use_dgi:
                            self.train_indices, self.indices_to_remove, self.model = KPGNN_model.initial_maintain(
                                train_i, i, self.data_split, self.metrics,
                                self.embedding_save_path, self.loss_fn, self.model)
                        else:
                            self.train_indices, self.indices_to_remove, self.model = KPGNN_model.initial_maintain(
                                train_i, i, self.data_split, self.metrics,
                                self.embedding_save_path, self.loss_fn, self.model,
                                self.loss_fn_dgi)

        data = SocialDataset(self.args.data_path, 0)
        g = dgl.DGLGraph(data.matrix)
        g.readonly()
        features = torch.FloatTensor(data.features)
        labels = torch.LongTensor(data.labels)

        predictions = []
        ground_truths = []
        self.detection_path = '../model/model_saved/kpgnn/detection_split/'
        os.makedirs(self.detection_path, exist_ok=True)

        test_indices = generateMasks(len(labels), self.data_split, 1, 0, 0.2, self.detection_path,
                                     num_indices_to_remove=0)

        g.ndata['features'] = features

        _, extract_features, extract_labels = extract_embeddings(g, self.model, len(labels), labels)

        # Extract labels
        test_indices = torch.load(self.detection_path + '/test_indices.pt')

        labels_true = extract_labels[test_indices]
        # Extract features
        X = extract_features[test_indices, :]
        assert labels_true.shape[0] == X.shape[0]
        n_test_tweets = X.shape[0]

        # Get the total number of classes
        n_classes = len(set(list(labels_true)))

        # kmeans clustering
        kmeans = KMeans(n_clusters=n_classes, random_state=0).fit(X)
        predictions = kmeans.labels_
        ground_truths = labels_true

        return predictions, ground_truths

    def evaluate(self, predictions, ground_truths):
        ars = metrics.adjusted_rand_score(ground_truths, predictions)

        # Calculate Adjusted Mutual Information (AMI)
        ami = metrics.adjusted_mutual_info_score(ground_truths, predictions)

        # Calculate Normalized Mutual Information (NMI)
        nmi = metrics.normalized_mutual_info_score(ground_truths, predictions)

        print(f"Model Adjusted Rand Index (ARI): {ars}")
        print(f"Model Adjusted Mutual Information (AMI): {ami}")
        print(f"Model Normalized Mutual Information (NMI): {nmi}")
        return ars, ami, nmi


class Preprocessor:
    def __init__(self, args, dataset):
        pass

    # generate_initial_features
    def generate_initial_features(self, dataset):
        save_path = './data/Event2012/kpgnn/'
        df = dataset

        os.makedirs(save_path, exist_ok=True)
        print("Data converted to dataframe.")
        print(df.shape)
        print(df.head(10))

        d_features = self.documents_to_features(df)
        print("Document features generated.")
        t_features = self.df_to_t_features(df)
        print("Time features generated.")
        combined_features = np.concatenate((d_features, t_features), axis=1)
        print("Concatenated document features and time features.")

        np.save(save_path + 'features_69612_0709_spacy_lg_zero_multiclasses_filtered.npy', combined_features)
        print("Initial features saved.")

    def documents_to_features(self, df):
        nlp = en_core_web_lg.load()
        print("df.filtered_words.head(10)", df.filtered_words.head(10))
        features = df.filtered_words.apply(lambda x: nlp(' '.join(x)).vector).values
        print("features.head(10)", features, "\n", "np.stack(features, axis=0)", np.stack(features, axis=0))
        return np.stack(features, axis=0)

    def extract_time_feature(self, t_str):
        t = datetime.fromisoformat(str(t_str))
        OLE_TIME_ZERO = datetime(1899, 12, 30)
        delta = t - OLE_TIME_ZERO
        return [(float(delta.days) / 100000.), (float(delta.seconds) / 86400)]  # 86,400 seconds in day

    # encode the times-tamps of all the messages in the dataframe
    def df_to_t_features(self, df):
        t_features = np.asarray([self.extract_time_feature(t_str) for t_str in df['created_at']])
        return t_features

    # custom_message_graph
    def custom_message_graph(self, dataset):
        save_path = '../model/model_saved/kpgnn/kpgnn_incremental_test/'
        '''
        if os.path.exists(save_path):
            pass
        else:
            os.mkdir(save_path)
        '''
        os.makedirs(save_path, exist_ok=True)

        df = dataset
        print("Data loaded.")

        # sort data by time
        df = df.sort_values(by='created_at').reset_index()

        # append date
        df['date'] = [d.date() for d in df['created_at']]

        # load features
        # the dimension of feature is 300 in this dataset
        f = np.load('../dataset/data/Event2012/kpgnn/features_69612_0709_spacy_lg_zero_multiclasses_filtered.npy')

        # generate test graphs, features, and labels
        message, data_split, all_graph_mins = self.construct_incremental_dataset(df, save_path, f, True)
        with open(save_path + "node_edge_statistics.txt", "w") as text_file:
            text_file.write(message)
        np.save(save_path + 'data_split.npy', np.asarray(data_split))
        print("Data split: ", data_split)
        np.save(save_path + 'all_graph_mins.npy', np.asarray(all_graph_mins))
        print("Time sepnt on heterogeneous -> homogeneous graph conversions: ", all_graph_mins)

    def construct_graph_from_df(self, df, G=None):
        if G is None:
            G = nx.Graph()
        for _, row in df.iterrows():
            tid = 't_' + str(row['tweet_id'])
            G.add_node(tid)
            G.nodes[tid]['tweet_id'] = True  # right-hand side value is irrelevant for the lookup

            user_ids = row['user_mentions']
            user_ids.append(row['user_id'])
            user_ids = ['u_' + str(each) for each in user_ids]
            # print(user_ids)
            G.add_nodes_from(user_ids)
            for each in user_ids:
                G.nodes[each]['user_id'] = True

            entities = row['entities']
            # entities = ['e_' + each for each in entities]
            # print(entities)
            G.add_nodes_from(entities)
            for each in entities:
                G.nodes[each]['entity'] = True

            words = row['sampled_words']
            words = ['w_' + each for each in words]
            # print(words)
            G.add_nodes_from(words)
            for each in words:
                G.nodes[each]['word'] = True

            edges = []
            edges += [(tid, each) for each in user_ids]
            edges += [(tid, each) for each in entities]
            edges += [(tid, each) for each in words]
            G.add_edges_from(edges)

        return G

    def networkx_to_dgl_graph(self, G, save_path=None):
        message = ''
        print('Start converting heterogeneous networkx graph to homogeneous dgl graph.')
        message += 'Start converting heterogeneous networkx graph to homogeneous dgl graph.\n'
        all_start = time.time()

        print('\tGetting a list of all nodes ...')
        message += '\tGetting a list of all nodes ...\n'
        start = time.time()
        all_nodes = list(G.nodes)
        mins = (time.time() - start) / 60
        print('\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        # print('All nodes: ', all_nodes)
        # print('Total number of nodes: ', len(all_nodes))

        print('\tGetting adjacency matrix ...')
        message += '\tGetting adjacency matrix ...\n'
        start = time.time()
        A = nx.to_numpy_array(G)  # Returns the graph adjacency matrix as a NumPy matrix.
        mins = (time.time() - start) / 60
        print('\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        # compute commuting matrices
        print('\tGetting lists of nodes of various types ...')
        message += '\tGetting lists of nodes of various types ...\n'
        start = time.time()
        tid_nodes = list(nx.get_node_attributes(G, 'tweet_id').keys())
        userid_nodes = list(nx.get_node_attributes(G, 'user_id').keys())
        word_nodes = list(nx.get_node_attributes(G, 'word').keys())
        entity_nodes = list(nx.get_node_attributes(G, 'entity').keys())
        del G
        mins = (time.time() - start) / 60
        print('\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        print('\tConverting node lists to index lists ...')
        message += '\tConverting node lists to index lists ...\n'
        start = time.time()
        #  find the index of target nodes in the list of all_nodes
        indices_tid = [all_nodes.index(x) for x in tid_nodes]
        indices_userid = [all_nodes.index(x) for x in userid_nodes]
        indices_word = [all_nodes.index(x) for x in word_nodes]
        indices_entity = [all_nodes.index(x) for x in entity_nodes]
        del tid_nodes
        del userid_nodes
        del word_nodes
        del entity_nodes
        mins = (time.time() - start) / 60
        print('\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        # ----------------------tweet-user-tweet----------------------
        print('\tStart constructing tweet-user-tweet commuting matrix ...')
        print('\t\t\tStart constructing tweet-user matrix ...')
        message += '\tStart constructing tweet-user-tweet commuting matrix ...\n\t\t\tStart constructing tweet-user ' \
                   'matrix ...\n '
        start = time.time()
        w_tid_userid = A[np.ix_(indices_tid, indices_userid)]
        #  return a N(indices_tid)*N(indices_userid) matrix, representing the weight of edges between tid and userid
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        # convert to scipy sparse matrix
        print('\t\t\tConverting to sparse matrix ...')
        message += '\t\t\tConverting to sparse matrix ...\n'
        start = time.time()
        s_w_tid_userid = sparse.csr_matrix(w_tid_userid)  # matrix compression
        del w_tid_userid
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        print('\t\t\tTransposing ...')
        message += '\t\t\tTransposing ...\n'
        start = time.time()
        s_w_userid_tid = s_w_tid_userid.transpose()
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        print('\t\t\tCalculating tweet-user * user-tweet ...')
        message += '\t\t\tCalculating tweet-user * user-tweet ...\n'
        start = time.time()
        s_m_tid_userid_tid = s_w_tid_userid * s_w_userid_tid  # homogeneous message graph
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        print('\t\t\tSaving ...')
        message += '\t\t\tSaving ...\n'
        start = time.time()
        if save_path is not None:
            sparse.save_npz(save_path + "s_m_tid_userid_tid.npz", s_m_tid_userid_tid)
            print("Sparse binary userid commuting matrix saved.")
            del s_m_tid_userid_tid
        del s_w_tid_userid
        del s_w_userid_tid
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        # ----------------------tweet-ent-tweet------------------------
        print('\tStart constructing tweet-ent-tweet commuting matrix ...')
        print('\t\t\tStart constructing tweet-ent matrix ...')
        message += '\tStart constructing tweet-ent-tweet commuting matrix ...\n\t\t\tStart constructing tweet-ent matrix ' \
                   '...\n '
        start = time.time()
        w_tid_entity = A[np.ix_(indices_tid, indices_entity)]
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        # convert to scipy sparse matrix
        print('\t\t\tConverting to sparse matrix ...')
        message += '\t\t\tConverting to sparse matrix ...\n'
        start = time.time()
        s_w_tid_entity = sparse.csr_matrix(w_tid_entity)
        del w_tid_entity
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        print('\t\t\tTransposing ...')
        message += '\t\t\tTransposing ...\n'
        start = time.time()
        s_w_entity_tid = s_w_tid_entity.transpose()
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        print('\t\t\tCalculating tweet-ent * ent-tweet ...')
        message += '\t\t\tCalculating tweet-ent * ent-tweet ...\n'
        start = time.time()
        s_m_tid_entity_tid = s_w_tid_entity * s_w_entity_tid
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        print('\t\t\tSaving ...')
        message += '\t\t\tSaving ...\n'
        start = time.time()
        if save_path is not None:
            sparse.save_npz(save_path + "s_m_tid_entity_tid.npz", s_m_tid_entity_tid)
            print("Sparse binary entity commuting matrix saved.")
            del s_m_tid_entity_tid
        del s_w_tid_entity
        del s_w_entity_tid
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        # ----------------------tweet-word-tweet----------------------
        print('\tStart constructing tweet-word-tweet commuting matrix ...')
        print('\t\t\tStart constructing tweet-word matrix ...')
        message += '\tStart constructing tweet-word-tweet commuting matrix ...\n\t\t\tStart constructing tweet-word ' \
                   'matrix ...\n '
        start = time.time()
        w_tid_word = A[np.ix_(indices_tid, indices_word)]
        del A
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        # convert to scipy sparse matrix
        print('\t\t\tConverting to sparse matrix ...')
        message += '\t\t\tConverting to sparse matrix ...\n'
        start = time.time()
        s_w_tid_word = sparse.csr_matrix(w_tid_word)
        del w_tid_word
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        print('\t\t\tTransposing ...')
        message += '\t\t\tTransposing ...\n'
        start = time.time()
        s_w_word_tid = s_w_tid_word.transpose()
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        print('\t\t\tCalculating tweet-word * word-tweet ...')
        message += '\t\t\tCalculating tweet-word * word-tweet ...\n'
        start = time.time()
        s_m_tid_word_tid = s_w_tid_word * s_w_word_tid
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        print('\t\t\tSaving ...')
        message += '\t\t\tSaving ...\n'
        start = time.time()
        if save_path is not None:
            sparse.save_npz(save_path + "s_m_tid_word_tid.npz", s_m_tid_word_tid)
            print("Sparse binary word commuting matrix saved.")
            del s_m_tid_word_tid
        del s_w_tid_word
        del s_w_word_tid
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'

        # ----------------------compute tweet-tweet adjacency matrix----------------------
        print('\tComputing tweet-tweet adjacency matrix ...')
        message += '\tComputing tweet-tweet adjacency matrix ...\n'
        start = time.time()
        if save_path is not None:
            s_m_tid_userid_tid = sparse.load_npz(save_path + "s_m_tid_userid_tid.npz")
            print("Sparse binary userid commuting matrix loaded.")
            s_m_tid_entity_tid = sparse.load_npz(save_path + "s_m_tid_entity_tid.npz")
            print("Sparse binary entity commuting matrix loaded.")
            s_m_tid_word_tid = sparse.load_npz(save_path + "s_m_tid_word_tid.npz")
            print("Sparse binary word commuting matrix loaded.")

        s_A_tid_tid = s_m_tid_userid_tid + s_m_tid_entity_tid
        del s_m_tid_userid_tid
        del s_m_tid_entity_tid
        s_bool_A_tid_tid = (s_A_tid_tid + s_m_tid_word_tid).astype('bool')  # confirm the connect between tweets
        del s_m_tid_word_tid
        del s_A_tid_tid
        mins = (time.time() - start) / 60
        print('\t\t\tDone. Time elapsed: ', mins, ' mins\n')
        message += '\t\t\tDone. Time elapsed: '
        message += str(mins)
        message += ' mins\n'
        all_mins = (time.time() - all_start) / 60
        print('\tOver all time elapsed: ', all_mins, ' mins\n')
        message += '\tOver all time elapsed: '
        message += str(all_mins)
        message += ' mins\n'

        if save_path is not None:
            sparse.save_npz(save_path + "s_bool_A_tid_tid.npz", s_bool_A_tid_tid)
            print("Sparse binary adjacency matrix saved.")
            s_bool_A_tid_tid = sparse.load_npz(save_path + "s_bool_A_tid_tid.npz")
            print("Sparse binary adjacency matrix loaded.")

        # create corresponding dgl graph
        G = dgl.DGLGraph(s_bool_A_tid_tid)
        print('We have %d nodes.' % G.number_of_nodes())
        print('We have %d edges.' % G.number_of_edges())
        print()
        message += 'We have '
        message += str(G.number_of_nodes())
        message += ' nodes.'
        message += 'We have '
        message += str(G.number_of_edges())
        message += ' edges.\n'

        return all_mins, message

    def construct_incremental_dataset(self, df, save_path, features, test=True):
        # If test equals true, construct the initial graph using test_ini_size tweets
        # and increment the graph by test_incr_size tweets each day
        test_ini_size = 500
        test_incr_size = 100

        # save data splits for training/validate/test mask generation
        data_split = []
        # save time spent for the heterogeneous -> homogeneous conversion of each graph
        all_graph_mins = []
        message = ""
        # extract distinct dates
        distinct_dates = df.date.unique()  # 2012-11-07
        # print("Distinct dates: ", distinct_dates)
        print("Number of distinct dates: ", len(distinct_dates))
        print()
        message += "Number of distinct dates: "
        message += str(len(distinct_dates))
        message += "\n"

        # split data by dates and construct graphs
        # first week -> initial graph (20254 tweets)
        print("Start constructing initial graph ...")
        message += "\nStart constructing initial graph ...\n"
        ini_df = df.loc[df['date'].isin(distinct_dates[:7])]  # find top 7 dates
        if test:
            ini_df = ini_df[:test_ini_size]  # top test_ini_size dates
        G = self.construct_graph_from_df(ini_df)
        path = save_path + '0/'
        if os.path.exists(path):
            pass
        else:
            os.mkdir(path)
        grap_mins, graph_message = self.networkx_to_dgl_graph(G, save_path=path)
        message += graph_message
        print("Initial graph saved")
        message += "Initial graph saved\n"
        # record the total number of tweets
        data_split.append(ini_df.shape[0])
        # record the time spent for graph conversion
        all_graph_mins.append(grap_mins)
        # extract and save the labels of corresponding tweets
        y = ini_df['event_id'].values
        y = [int(each) for each in y]
        np.save(path + 'labels.npy', np.asarray(y))
        print("Labels saved.")
        message += "Labels saved.\n"
        # extract and save the features of corresponding tweets
        indices = ini_df['index'].values.tolist()
        x = features[indices, :]
        np.save(path + 'features.npy', x)
        print("Features saved.")
        message += "Features saved.\n\n"

        # subsequent days -> insert tweets day by day (skip the last day because it only contains one tweet)
        for i in range(7, len(distinct_dates) - 1):
            print("Start constructing graph ", str(i - 6), " ...")
            message += "\nStart constructing graph "
            message += str(i - 6)
            message += " ...\n"
            incr_df = df.loc[df['date'] == distinct_dates[i]]
            if test:
                incr_df = incr_df[:test_incr_size]

            # All/Relevant Message Strategy: keeping all the messages when constructing the graphs 
            # (for the Relevant Message Strategy, the unrelated messages will be removed from the graph later on).
            # G = construct_graph_from_df(incr_df, G) 

            # Latest Message Strategy: construct graph using only the data of the day
            G = self.construct_graph_from_df(incr_df)

            path = save_path + str(i - 6) + '/'
            if os.path.exists(path):
                pass
            else:
                os.mkdir(path)

            grap_mins, graph_message = self.networkx_to_dgl_graph(G, save_path=path)
            message += graph_message
            print("Graph ", str(i - 6), " saved")
            message += "Graph "
            message += str(i - 6)
            message += " saved\n"
            # record the total number of tweets
            data_split.append(incr_df.shape[0])
            # record the time spent for graph conversion
            all_graph_mins.append(grap_mins)
            # extract and save the labels of corresponding tweets
            # y = np.concatenate([y, incr_df['event_id'].values], axis = 0)
            y = [int(each) for each in incr_df['event_id'].values]
            np.save(path + 'labels.npy', y)
            print("Labels saved.")
            message += "Labels saved.\n"
            # extract and save the features of corresponding tweets
            indices = incr_df['index'].values.tolist()
            x = features[indices, :]
            # x = np.concatenate([x, x_incr], axis = 0)
            np.save(path + 'features.npy', x)
            print("Features saved.")
            message += "Features saved.\n"

        return message, data_split, all_graph_mins


class KPGNN_model():
    def __init__(self):
        pass

    # Inference(prediction)
    def infer(train_i, i, data_split, metrics, embedding_save_path, loss_fn, train_indices=None, model=None,
              loss_fn_dgi=None, indices_to_remove=[]):
        args = args_define.args

        save_path_i = embedding_save_path + '/block_' + str(i)
        if not os.path.isdir(save_path_i):
            os.mkdir(save_path_i)

        data = SocialDataset(args.data_path, i)
        features = torch.FloatTensor(data.features)
        labels = torch.LongTensor(data.labels)
        print("labels1:", labels)
        in_feats = features.shape[1]

        g = dgl.graph((data.matrix.row, data.matrix.col))
        num_isolated_nodes = graph_statistics(g, save_path_i)

        if args.remove_obsolete == 1:
            if ((args.resume_path is not None) and (not args.resume_current) and (i == args.resume_point + 1) and (
                    i > args.window_size)) \
                    or (indices_to_remove == [] and i > args.window_size):
                temp_i = max(((i - 1) // args.window_size) * args.window_size, 0)
                indices_to_remove = np.load(
                    embedding_save_path + '/block_' + str(temp_i) + '/indices_to_remove.npy').tolist()

            if indices_to_remove != []:
                data.remove_obsolete_nodes(indices_to_remove)
                features = torch.FloatTensor(data.features)
                labels = torch.LongTensor(data.labels)
                print("labels2:", labels)
                g = dgl.graph((data.matrix.row, data.matrix.col))
                num_isolated_nodes = graph_statistics(g, save_path_i)

        if args.mask_path is None:
            mask_path = save_path_i + '/masks'
            if not os.path.isdir(mask_path):
                os.mkdir(mask_path)
            test_indices = generateMasks(len(labels), data_split, train_i, i, args.validation_percent, mask_path,
                                         len(indices_to_remove))
        else:
            test_indices = torch.load(args.mask_path + '/block_' + str(i) + '/masks/test_indices.pt')

        if args.use_cuda:
            features, labels = features.cuda(), labels.cuda()
            print("labels3:", labels)
            test_indices = test_indices.cuda()

        g.ndata['features'] = features

        if (args.resume_path is not None) and (not args.resume_current) and (i == args.resume_point + 1):
            if args.use_dgi:
                model = DGI(in_feats, args.hidden_dim, args.out_dim, args.num_heads, args.use_residual)
            else:
                model = GAT(in_feats, args.hidden_dim, args.out_dim, args.num_heads, args.use_residual)

            if args.use_cuda:
                model.cuda()

            model_path = embedding_save_path + '/block_' + str(args.resume_point) + '/models/best.pt'
            model.load_state_dict(torch.load(model_path))
            print("Resumed model from the previous block.")

            args.resume_path = None

        if train_indices is None:
            if args.remove_obsolete == 0 or args.remove_obsolete == 1:
                temp_i = max(((i - 1) // args.window_size) * args.window_size, 0)
                train_indices = torch.load(embedding_save_path + '/block_' + str(temp_i) + '/masks/train_indices.pt')
            else:
                if args.n_infer_epochs != 0:
                    print(
                        "==================================\n'continue training then predict' is unimplemented under remove_obsolete mode 2, will skip infer epochs.\n===================================\n")
                    args.n_infer_epochs = 0

        all_test_nmi = []
        time_predict = []

        message = "\n------------ Directly predict on block " + str(i) + " ------------\n"
        print(message)
        with open(save_path_i + '/log.txt', 'a') as f:
            f.write(message)
        start = time.time()

        extract_nids, extract_features, extract_labels = extract_embeddings(g, model, len(labels), labels)
        test_nmi = evaluate_model(extract_features, extract_labels, test_indices, -1, num_isolated_nodes, save_path_i,
                                  False)
        seconds_spent = time.time() - start
        message = '\nDirect prediction took {:.2f} seconds'.format(seconds_spent)
        print(message)
        with open(save_path_i + '/log.txt', 'a') as f:
            f.write(message)
        all_test_nmi.append(test_nmi)
        time_predict.append(seconds_spent)
        np.save(save_path_i + '/time_predict.npy', np.asarray(time_predict))

        optimizer = optim.Adam(model.parameters(), lr=args.lr, weight_decay=1e-4)

        if args.n_infer_epochs != 0:
            message = "\n------------ Continue training then predict on block " + str(i) + " ------------\n"
            print(message)
            with open(save_path_i + '/log.txt', 'a') as f:
                f.write(message)
        seconds_infer_batches = []
        mins_infer_epochs = []

        sampler = MultiLayerNeighborSampler([args.n_neighbors] * 2)
        dataloader = NodeDataLoader(
            g, train_indices, sampler,
            batch_size=args.batch_size,
            shuffle=True,
            drop_last=False,
            num_workers=4)

        for epoch in range(args.n_infer_epochs):
            start_epoch = time.time()
            losses = []
            total_loss = 0
            if args.use_dgi:
                losses_triplet = []
                losses_dgi = []
            for metric in metrics:
                metric.reset()

            for batch_id, (input_nodes, output_nodes, blocks) in enumerate(dataloader):
                start_batch = time.time()
                batch_features = blocks[0].srcdata['features']
                model.train()

                if args.use_dgi:
                    pred, ret = model(blocks, batch_features)
                else:
                    pred = model(blocks, batch_features)

                batch_labels = labels[output_nodes]
                loss_outputs = loss_fn(pred, batch_labels)
                loss = loss_outputs[0] if isinstance(loss_outputs, (tuple, list)) else loss_outputs
                if args.use_dgi:
                    n_samples = len(output_nodes)
                    lbl_1 = torch.ones(n_samples)
                    lbl_2 = torch.zeros(n_samples)
                    lbl = torch.cat((lbl_1, lbl_2), 0)
                    if args.use_cuda:
                        lbl = lbl.cuda()
                    losses_triplet.append(loss.item())
                    loss_dgi = loss_fn_dgi(ret, lbl)
                    losses_dgi.append(loss_dgi.item())
                    loss += loss_dgi
                    losses.append(loss.item())
                else:
                    losses.append(loss.item())
                total_loss += loss.item()

                for metric in metrics:
                    metric(pred, batch_labels, loss_outputs)

                if batch_id % args.log_interval == 0:
                    message = 'Train: [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                        batch_id * args.batch_size, train_indices.shape[0],
                        100. * batch_id / (train_indices.shape[0] // args.batch_size), np.mean(losses))
                    if args.use_dgi:
                        message += '\tLoss_triplet: {:.6f}'.format(np.mean(losses_triplet))
                        message += '\tLoss_dgi: {:.6f}'.format(np.mean(losses_dgi))
                    for metric in metrics:
                        message += '\t{}: {:.4f}'.format(metric.name(), metric.value())
                    print(message)
                    with open(save_path_i + '/log.txt', 'a') as f:
                        f.write(message)
                    losses = []

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                batch_seconds_spent = time.time() - start_batch
                seconds_infer_batches.append(batch_seconds_spent)

            total_loss /= (batch_id + 1)
            message = 'Epoch: {}/{}. Average loss: {:.4f}'.format(epoch + 1, args.n_infer_epochs, total_loss)
            for metric in metrics:
                message += '\t{}: {:.4f}'.format(metric.name(), metric.value())
            mins_spent = (time.time() - start_epoch) / 60
            message += '\nThis epoch took {:.2f} mins'.format(mins_spent)
            message += '\n'
            print(message)
            with open(save_path_i + '/log.txt', 'a') as f:
                f.write(message)
            mins_infer_epochs.append(mins_spent)

            # 验证
            extract_nids, extract_features, extract_labels = extract_embeddings(g, model, len(labels), labels)
            test_nmi = evaluate_model(extract_features, extract_labels, test_indices, epoch, num_isolated_nodes,
                                      save_path_i, False)
            all_test_nmi.append(test_nmi)
            # end one epoch

        # Save model (fine-tuned from the above continue training process)
        model_path = save_path_i + '/models'
        os.mkdir(model_path)
        p = model_path + '/best.pt'
        torch.save(model.state_dict(), p)
        print('Model saved.')

        # Save all test nmi
        np.save(save_path_i + '/all_test_nmi.npy', np.asarray(all_test_nmi))
        print('Saved all test nmi.')
        # Save time spent on epochs
        np.save(save_path_i + '/mins_infer_epochs.npy', np.asarray(mins_infer_epochs))
        print('Saved mins_infer_epochs.')
        # Save time spent on batches
        np.save(save_path_i + '/seconds_infer_batches.npy', np.asarray(seconds_infer_batches))
        print('Saved seconds_infer_batches.')

        return model

    # Train on initial/maintenance graphs, t == 0 or t % window_size == 0 in this paper
    def initial_maintain(train_i, i, data_split, metrics, embedding_save_path, loss_fn, model=None, loss_fn_dgi=None):
        args = args_define().args
        # make dir for graph i
        save_path_i = embedding_save_path + '/block_' + str(i)
        if not os.path.isdir(save_path_i):
            os.mkdir(save_path_i)

        # load data
        data = SocialDataset(args.data_path, i)
        features = torch.FloatTensor(data.features)
        labels = torch.LongTensor(data.labels)
        in_feats = features.shape[1]  # feature dimension

        # Construct graph that contains message blocks 0, ..., i if remove_obsolete = 0 or 1; graph that only contains message block i if remove_obsolete = 2
        g = dgl.DGLGraph(data.matrix)
        num_isolated_nodes = graph_statistics(g, save_path_i)

        # if remove_obsolete is mode 1, resume or generate indices_to_remove, then remove obsolete nodes from the graph
        if args.remove_obsolete == 1:

            # Resume indices_to_remove from the current block
            if (args.resume_path is not None) and args.resume_current and (i == args.resume_point) and (i != 0):
                indices_to_remove = np.load(save_path_i + '/indices_to_remove.npy').tolist()

            elif i == 0:  # generate empty indices_to_remove for initial block
                indices_to_remove = []
                # save indices_to_remove
                np.save(save_path_i + '/indices_to_remove.npy', np.asarray(indices_to_remove))

            #  update graph
            else:  # generate indices_to_remove for maintenance block
                # get the indices of all training nodes
                num_all_train_nodes = np.sum(data_split[:i + 1])
                all_train_indices = np.arange(0, num_all_train_nodes).tolist()
                # get the number of old training nodes added before this maintenance
                num_old_train_nodes = np.sum(data_split[:i + 1 - args.window_size])
                # indices_to_keep: indices of nodes that are connected to the new training nodes added at this maintenance
                # (include the indices of the new training nodes)
                indices_to_keep = list(set(data.matrix.indices[data.matrix.indptr[num_old_train_nodes]:]))
                # indices_to_remove is the difference between the indices of all training nodes and indices_to_keep
                indices_to_remove = list(set(all_train_indices) - set(indices_to_keep))
                # save indices_to_remove
                np.save(save_path_i + '/indices_to_remove.npy', np.asarray(indices_to_remove))

            if indices_to_remove != []:
                # remove obsolete nodes from the graph
                data.remove_obsolete_nodes(indices_to_remove)
                features = torch.FloatTensor(data.features)
                labels = torch.LongTensor(data.labels)
                # Reconstruct graph
                g = dgl.DGLGraph(data.matrix)  # graph that contains tweet blocks 0, ..., i
                num_isolated_nodes = graph_statistics(g, save_path_i)

        else:

            indices_to_remove = []

        # generate or load training/validate/test masks
        if (args.resume_path is not None) and args.resume_current and (
                i == args.resume_point):  # Resume masks from the current block

            train_indices = torch.load(save_path_i + '/masks/train_indices.pt')
            validation_indices = torch.load(save_path_i + '/masks/validation_indices.pt')
        if args.mask_path is None:

            mask_path = save_path_i + '/masks'
            if not os.path.isdir(mask_path):
                os.mkdir(mask_path)
            train_indices, validation_indices = generateMasks(len(labels), data_split, train_i, i,
                                                              args.validation_percent,
                                                              mask_path, len(indices_to_remove))

        else:
            train_indices = torch.load(args.mask_path + '/block_' + str(i) + '/masks/train_indices.pt')
            validation_indices = torch.load(args.mask_path + '/block_' + str(i) + '/masks/validation_indices.pt')

        # Suppress warning
        g.set_n_initializer(dgl.init.zero_initializer)

        if args.use_cuda:
            features, labels = features.cuda(), labels.cuda()
            train_indices, validation_indices = train_indices.cuda(), validation_indices.cuda()

        g.ndata['features'] = features

        if (args.resume_path is not None) and args.resume_current and (
                i == args.resume_point):  # Resume model from the current block

            # Declare model
            if args.use_dgi:
                model = DGI(in_feats, args.hidden_dim, args.out_dim, args.num_heads, args.use_residual)
            else:
                model = GAT(in_feats, args.hidden_dim, args.out_dim, args.num_heads, args.use_residual)

            if args.use_cuda:
                model.cuda()

            # Load model from resume_point
            model_path = embedding_save_path + '/block_' + str(args.resume_point) + '/models/best.pt'
            model.load_state_dict(torch.load(model_path))
            print("Resumed model from the current block.")

            # Use resume_path as a flag
            args.resume_path = None

        elif model is None:  # Construct the initial model
            # Declare model
            if args.use_dgi:
                model = DGI(in_feats, args.hidden_dim, args.out_dim, args.num_heads, args.use_residual)
            else:
                model = GAT(in_feats, args.hidden_dim, args.out_dim, args.num_heads, args.use_residual)

            if args.use_cuda:
                model.cuda()

        # Optimizer
        optimizer = optim.Adam(model.parameters(), lr=args.lr, weight_decay=1e-4)

        # Start training
        message = "\n------------ Start initial training / maintaining using blocks 0 to " + str(i) + " ------------\n"
        print(message)
        with open(save_path_i + '/log.txt', 'a') as f:
            f.write(message)
        # record the highest validation nmi ever got for early stopping
        best_vali_nmi = 1e-9
        best_epoch = 0
        wait = 0
        # record validation nmi of all epochs before early stop
        all_vali_nmi = []
        # record the time spent in seconds on each batch of all training/maintaining epochs
        seconds_train_batches = []
        # record the time spent in mins on each epoch
        mins_train_epochs = []
        g.readonly()

        sampler = MultiLayerNeighborSampler([args.n_neighbors] * 2)  # args.n_hops 应该是 2

        # 创建 DataLoader
        dataloader = NodeDataLoader(
            g, train_indices, sampler,
            batch_size=args.batch_size,
            shuffle=True,
            drop_last=False,
            num_workers=4)  # 设置为适当的 num_workers

        for epoch in range(args.n_epochs):
            start_epoch = time.time()
            model.train()
            total_loss = 0
            losses = []
            if args.use_dgi:
                losses_triplet = []
                losses_dgi = []

            for metric in metrics:
                metric.reset()

            for batch_id, (input_nodes, output_nodes, blocks) in enumerate(dataloader):
                start_batch = time.time()

                blocks = [block.int().to(torch.device('cuda' if args.use_cuda else 'cpu')) for block in blocks]
                batch_features = blocks[0].srcdata['features']
                batch_labels = labels[output_nodes]

                # forward
                if args.use_dgi:
                    pred, ret = model(blocks, batch_features)
                else:
                    blocks[0].srcdata['h'] = batch_features
                    # print(blocks[0].srcdata)

                    pred = model(blocks, batch_features)

                loss_outputs = loss_fn(pred, batch_labels)
                loss = loss_outputs[0] if isinstance(loss_outputs, (tuple, list)) else loss_outputs

                if args.use_dgi:
                    n_samples = len(output_nodes)
                    lbl_1 = torch.ones(n_samples)
                    lbl_2 = torch.zeros(n_samples)
                    lbl = torch.cat((lbl_1, lbl_2), 0)
                    if args.use_cuda:
                        lbl = lbl.cuda()
                    losses_triplet.append(loss.item())
                    loss_dgi = loss_fn_dgi(ret, lbl)
                    losses_dgi.append(loss_dgi.item())
                    loss += loss_dgi
                    losses.append(loss.item())
                else:
                    losses.append(loss.item())

                total_loss += loss.item()

                for metric in metrics:
                    metric(pred, batch_labels, loss_outputs)

                if batch_id % args.log_interval == 0:
                    message = 'Train: [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                        batch_id * args.batch_size, train_indices.shape[0],
                        100. * batch_id / (len(dataloader)), np.mean(losses))
                    if args.use_dgi:
                        message += '\tLoss_triplet: {:.6f}'.format(np.mean(losses_triplet))
                        message += '\tLoss_dgi: {:.6f}'.format(np.mean(losses_dgi))
                    for metric in metrics:
                        message += '\t{}: {:.4f}'.format(metric.name(), metric.value())
                    print(message)
                    with open(save_path_i + '/log.txt', 'a') as f:
                        f.write(message)
                    losses = []

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                batch_seconds_spent = time.time() - start_batch
                seconds_train_batches.append(batch_seconds_spent)

            total_loss /= (batch_id + 1)
            message = 'Epoch: {}/{}. Average loss: {:.4f}'.format(epoch + 1, args.n_epochs, total_loss)
            for metric in metrics:
                message += '\t{}: {:.4f}'.format(metric.name(), metric.value())
            mins_spent = (time.time() - start_epoch) / 60
            message += '\nThis epoch took {:.2f} mins'.format(mins_spent)
            message += '\n'
            print(message)
            with open(save_path_i + '/log.txt', 'a') as f:
                f.write(message)
            mins_train_epochs.append(mins_spent)

            # Validation
            # Infer the representations of all tweets
            g.readonly()
            extract_nids, extract_features, extract_labels = extract_embeddings(g, model, len(labels), labels)
            # Save the representations of all tweets
            # save_embeddings(extract_nids, extract_features, extract_labels, extract_train_tags, save_path_i, epoch)
            # Evaluate the model: conduct kMeans clustering on the validation and report NMI
            validation_nmi = evaluate_model(extract_features, extract_labels, validation_indices, epoch,
                                            num_isolated_nodes,
                                            save_path_i, True)
            all_vali_nmi.append(validation_nmi)

            # Early stop
            if validation_nmi > best_vali_nmi:
                best_vali_nmi = validation_nmi
                best_epoch = epoch
                wait = 0
                # Save model
                model_path = save_path_i + '/models'
                if (epoch == 0) and (not os.path.isdir(model_path)):
                    os.mkdir(model_path)
                p = model_path + '/best.pt'
                torch.save(model.state_dict(), p)
                print('Best model saved after epoch ', str(epoch))
            else:
                wait += 1
            if wait == args.patience:
                print('Saved all_mins_spent')
                print('Early stopping at epoch ', str(epoch))
                print('Best model was at epoch ', str(best_epoch))
                break
            # end one epoch

        # Save all validation nmi
        np.save(save_path_i + '/all_vali_nmi.npy', np.asarray(all_vali_nmi))
        # Save time spent on epochs
        np.save(save_path_i + '/mins_train_epochs.npy', np.asarray(mins_train_epochs))
        print('Saved mins_train_epochs.')
        # Save time spent on batches
        np.save(save_path_i + '/seconds_train_batches.npy', np.asarray(seconds_train_batches))
        print('Saved seconds_train_batches.')

        # Load the best model of the current block
        best_model_path = save_path_i + '/models/best.pt'
        model.load_state_dict(torch.load(best_model_path))
        print("Best model loaded.")

        if args.remove_obsolete == 2:
            return None, indices_to_remove, model
        return train_indices, indices_to_remove, model


def graph_statistics(G, save_path):
    message = '\nGraph statistics:\n'

    num_nodes = G.number_of_nodes()
    num_edges = G.number_of_edges()
    ave_degree = (num_edges / 2) // num_nodes
    in_degrees = G.in_degrees()
    isolated_nodes = torch.zeros([in_degrees.size()[0]], dtype=torch.long)
    isolated_nodes = (in_degrees == isolated_nodes)
    torch.save(isolated_nodes, save_path + '/isolated_nodes.pt')
    num_isolated_nodes = torch.sum(isolated_nodes).item()

    message += 'We have ' + str(num_nodes) + ' nodes.\n'
    message += 'We have ' + str(num_edges / 2) + ' in-edges.\n'
    message += 'Average degree: ' + str(ave_degree) + '\n'
    message += 'Number of isolated nodes: ' + str(num_isolated_nodes) + '\n'
    print(message)
    with open(save_path + "/graph_statistics.txt", "a") as f:
        f.write(message)

    return num_isolated_nodes


def generateMasks(length, data_split, train_i, i, validation_percent=0.2, save_path=None, num_indices_to_remove=0):
    """
        Intro:
        This function generates train and validation indices for initial/maintenance epochs and test indices for inference(prediction) epochs
        If remove_obsolete mode 0 or 1:
        For initial/maintenance epochs:
        - The first (train_i + 1) blocks (blocks 0, ..., train_i) are used as training set (with explicit labels)
        - Randomly sample validation_percent of the training indices as validation indices
        For inference(prediction) epochs:
        - The (i + 1)th block (block i) is used as test set
        Note that other blocks (block train_i + 1, ..., i - 1) are also in the graph (without explicit labels, only their features and structural info are leveraged)
        If remove_obsolete mode 2:
        For initial/maintenance epochs:
        - The (i + 1) = (train_i + 1)th block (block train_i = i) is used as training set (with explicit labels)
        - Randomly sample validation_percent of the training indices as validation indices
        For inference(prediction) epochs:
        - The (i + 1)th block (block i) is used as test set

        :param length: the length of label list
        :param data_split: loaded splited data (generated in custom_message_graph.py)
        :param train_i, i: flag, indicating for initial/maintenance stage if train_i == i and inference stage for others
        :param validation_percent: the percent of validation data occupied in whole dataset
        :param save_path: path to save data
        :param num_indices_to_remove: number of indices ought to be removed

        :returns train indices, validation indices or test indices
    """
    args = args_define().args
    if args.remove_obsolete == 0 or args.remove_obsolete == 1:  # remove_obsolete mode 0 or 1
        # verify total number of nodes
        assert length == (np.sum(data_split[:i + 1]) - num_indices_to_remove)

        # If is in initial/maintenance epochs, generate train and validation indices
        if train_i == i:
            # randomly shuffle the training indices
            train_length = np.sum(data_split[:train_i + 1])
            train_length -= num_indices_to_remove
            train_indices = torch.randperm(int(train_length))
            # get total number of validation indices
            n_validation_samples = int(train_length * validation_percent)
            # sample n_validation_samples validation indices and use the rest as training indices
            validation_indices = train_indices[:n_validation_samples]
            train_indices = train_indices[n_validation_samples:]
            if save_path is not None:
                torch.save(validation_indices, save_path + '/validation_indices.pt')
                torch.save(train_indices, save_path + '/train_indices.pt')
                validation_indices = torch.load(save_path + '/validation_indices.pt')
                train_indices = torch.load(save_path + '/train_indices.pt')
            return train_indices, validation_indices
        # If the process is in inference(prediction) epochs, generate test indices
        else:
            test_indices = torch.range(0, (data_split[i] - 1), dtype=torch.long)
            test_indices += (np.sum(data_split[:i]) - num_indices_to_remove)
            if save_path is not None:
                torch.save(test_indices, save_path + '/test_indices.pt')
                test_indices = torch.load(save_path + '/test_indices.pt')
            return test_indices

    else:  # remove_obsolete mode 2
        # verify total number of nodes
        assert length == data_split[i]

        # If is in initial/maintenance epochs, generate train and validation indices
        if train_i == i:
            # randomly shuffle the graph indices
            train_indices = torch.randperm(length)
            # get total number of validation indices
            n_validation_samples = int(length * validation_percent)
            # sample n_validation_samples validation indices and use the rest as training indices
            validation_indices = train_indices[:n_validation_samples]
            train_indices = train_indices[n_validation_samples:]
            if save_path is not None:
                torch.save(validation_indices, save_path +
                           '/validation_indices.pt')
                torch.save(train_indices, save_path + '/train_indices.pt')
                validation_indices = torch.load(
                    save_path + '/validation_indices.pt')
                train_indices = torch.load(save_path + '/train_indices.pt')
            return train_indices, validation_indices
        # If is in inference(prediction) epochs, generate test indices
        else:
            test_indices = torch.range(
                0, (data_split[i] - 1), dtype=torch.long)
            if save_path is not None:
                torch.save(test_indices, save_path + '/test_indices.pt')
                test_indices = torch.load(save_path + '/test_indices.pt')
            return test_indices


# Compute the representations of all the nodes in g using model
def extract_embeddings(g, model, num_all_samples, labels):
    args = args_define().args
    sampler = MultiLayerNeighborSampler([1000, 1000])
    dataloader = NodeDataLoader(
        g, torch.arange(g.num_nodes()), sampler,
        batch_size=num_all_samples,
        shuffle=False,
        drop_last=False,
        num_workers=4)  # 设置合适的 num_workers

    with torch.no_grad():
        model.eval()
        for batch_id, (input_nodes, output_nodes, blocks) in enumerate(dataloader):
            batch_features = blocks[0].srcdata['features']

            if args.use_dgi:
                extract_features, _ = model(blocks, batch_features)
            else:
                extract_features = model(blocks, batch_features)

            extract_nids = output_nodes.to(device=extract_features.device, dtype=torch.long)
            extract_labels = labels[extract_nids]

            assert batch_id == 0
            extract_nids = extract_nids.cpu().numpy()
            extract_features = extract_features.cpu().numpy()
            extract_labels = extract_labels.cpu().numpy()

        A = np.arange(num_all_samples)
        assert (A == extract_nids).all()

    return extract_nids, extract_features, extract_labels


def save_embeddings(extract_nids, extract_features, extract_labels, extract_train_tags, path, counter):
    np.savetxt(path + '/features_' + str(counter) + '.tsv', extract_features, delimiter='\t')
    np.savetxt(path + '/labels_' + str(counter) + '.tsv', extract_labels, fmt='%i', delimiter='\t')
    with open(path + '/labels_tags_' + str(counter) + '.tsv', 'w') as f:
        f.write('label\tmessage_id\ttrain_tag\n')
        for (label, mid, train_tag) in zip(extract_labels, extract_nids, extract_train_tags):
            f.write("%s\t%s\t%s\n" % (label, mid, train_tag))
    print("Embeddings after inference epoch " + str(counter) + " saved.")
    print()


def intersection(lst1, lst2):
    lst3 = [value for value in lst1 if value in lst2]
    return lst3


def run_kmeans(extract_features, extract_labels, indices, isoPath=None):
    # Extract the features and labels of the test tweets
    indices = indices.cpu().detach().numpy()

    if isoPath is not None:
        # Remove isolated points
        temp = torch.load(isoPath)
        temp = temp.cpu().detach().numpy()
        non_isolated_index = list(np.where(temp != 1)[0])
        indices = intersection(indices, non_isolated_index)

    # Extract labels
    labels_true = extract_labels[indices]
    # Extract features
    X = extract_features[indices, :]
    assert labels_true.shape[0] == X.shape[0]
    n_test_tweets = X.shape[0]

    # Get the total number of classes     
    n_classes = len(set(list(labels_true)))

    # k-means clustering
    kmeans = KMeans(n_clusters=n_classes, random_state=0).fit(X)
    labels = kmeans.labels_
    print("n_classes:", n_classes)
    print("labels_true list:", labels_true.tolist())
    print("labels_pred list:", labels.tolist())

    nmi = metrics.normalized_mutual_info_score(labels_true, labels)

    # Return number of test tweets, number of classes covered by the test tweets, and kMeans cluatering NMI
    return (n_test_tweets, n_classes, nmi)


def evaluate_model(extract_features, extract_labels, indices, epoch, num_isolated_nodes, save_path, is_validation=True):
    message = ''
    message += '\nEpoch '
    message += str(epoch)
    message += '\n'

    # with isolated nodes
    n_tweets, n_classes, nmi = run_kmeans(extract_features, extract_labels, indices)
    if is_validation:
        mode = 'validation'
    else:
        mode = 'test'
    message += '\tNumber of ' + mode + ' tweets: '
    message += str(n_tweets)
    message += '\n\tNumber of classes covered by ' + mode + ' tweets: '
    message += str(n_classes)
    message += '\n\t' + mode + ' NMI: '
    message += str(nmi)
    if num_isolated_nodes != 0:
        # without isolated nodes
        message += '\n\tWithout isolated nodes:'
        n_tweets, n_classes, nmi = run_kmeans(extract_features, extract_labels, indices,
                                              save_path + '/isolated_nodes.pt')
        message += '\tNumber of ' + mode + ' tweets: '
        message += str(n_tweets)
        message += '\n\tNumber of classes covered by ' + mode + ' tweets: '
        message += str(n_classes)
        message += '\n\t' + mode + ' NMI: '
        message += str(nmi)
    message += '\n'

    with open(save_path + '/evaluate.txt', 'a') as f:
        f.write(message)
    print(message)

    return nmi


class Metric:
    def __init__(self):
        pass

    def __call__(self, outputs, target, loss):
        raise NotImplementedError

    def reset(self):
        raise NotImplementedError

    def value(self):
        raise NotImplementedError

    def name(self):
        raise NotImplementedError


class AccumulatedAccuracyMetric(Metric):
    """
    Works with classification model
    """

    def __init__(self):
        self.correct = 0
        self.total = 0

    def __call__(self, outputs, target, loss):
        pred = outputs[0].data.max(1, keepdim=True)[1]
        self.correct += pred.eq(target[0].data.view_as(pred)).cpu().sum()
        self.total += target[0].size(0)
        return self.value()

    def reset(self):
        self.correct = 0
        self.total = 0

    def value(self):
        return 100 * float(self.correct) / self.total

    def name(self):
        return 'Accuracy'


class AverageNonzeroTripletsMetric(Metric):
    '''
    Counts average number of nonzero triplets found in minibatches
    '''

    def __init__(self):
        self.values = []

    def __call__(self, outputs, target, loss):
        self.values.append(loss[1])
        return self.value()

    def reset(self):
        self.values = []

    def value(self):
        return np.mean(self.values)

    def name(self):
        return 'Average nonzero triplets'


# layer
class GATLayer(nn.Module):
    def __init__(self, in_dim, out_dim, use_residual=False):
        super(GATLayer, self).__init__()
        self.fc = nn.Linear(in_dim, out_dim, bias=False)
        self.attn_fc = nn.Linear(2 * out_dim, 1, bias=False)
        self.use_residual = use_residual
        self.reset_parameters()

    def reset_parameters(self):
        """Reinitialize learnable parameters."""
        gain = nn.init.calculate_gain('relu')
        nn.init.xavier_normal_(self.fc.weight, gain=gain)
        nn.init.xavier_normal_(self.attn_fc.weight, gain=gain)

    def edge_attention(self, edges):
        # Edge UDF for equation (2)
        z2 = torch.cat([edges.src['z'], edges.dst['z']], dim=1)
        a = self.attn_fc(z2)
        return {'e': F.leaky_relu(a)}

    def message_func(self, edges):
        # Message UDF for equation (3) & (4)
        return {'z': edges.src['z'], 'e': edges.data['e']}

    def reduce_func(self, nodes):
        # Reduce UDF for equation (3) & (4)
        alpha = F.softmax(nodes.mailbox['e'], dim=1)
        h = torch.sum(alpha * nodes.mailbox['z'], dim=1)
        return {'h': h}

    def forward(self, block):
        h = block.srcdata['h']
        z = self.fc(h)
        block.srcdata['z'] = z
        block.dstdata['z'] = z[:block.num_dst_nodes()]  # 确保 dstdata['z'] 也被正确设置

        block.apply_edges(self.edge_attention)
        block.update_all(self.message_func, self.reduce_func)

        if self.use_residual:
            return z[:block.num_dst_nodes()] + block.dstdata['h']
        else:
            return block.dstdata['h']


class MultiHeadGATLayer(nn.Module):
    def __init__(self, in_dim, out_dim, num_heads, merge='cat', use_residual=False):
        super(MultiHeadGATLayer, self).__init__()
        self.heads = nn.ModuleList()
        for i in range(num_heads):
            self.heads.append(GATLayer(in_dim, out_dim, use_residual))
        self.merge = merge

    def forward(self, block):
        head_outs = [attn_head(block) for attn_head in self.heads]
        if self.merge == 'cat':
            return torch.cat(head_outs, dim=1)
        else:
            return torch.mean(torch.stack(head_outs), dim=0)


class GAT(nn.Module):
    def __init__(self, in_dim, hidden_dim, out_dim, num_heads, use_residual=False):
        super(GAT, self).__init__()
        self.layer1 = MultiHeadGATLayer(in_dim, hidden_dim, num_heads, 'cat', use_residual)
        self.layer2 = MultiHeadGATLayer(hidden_dim * num_heads, out_dim, 1, 'cat', use_residual)

    def forward(self, blocks, features):
        blocks[0].srcdata['h'] = features

        h = self.layer1(blocks[0])
        h = F.elu(h)
        blocks[1].srcdata['h'] = h

        h = self.layer2(blocks[1])

        return h


# Applies an average on seq, of shape (nodes, features)
class AvgReadout(nn.Module):
    def __init__(self):
        super(AvgReadout, self).__init__()

    def forward(self, seq):
        return torch.mean(seq, 0)


class Discriminator(nn.Module):
    def __init__(self, n_h):
        super(Discriminator, self).__init__()
        self.f_k = nn.Bilinear(n_h, n_h, 1)
        for m in self.modules():
            self.weights_init(m)

    def weights_init(self, m):
        if isinstance(m, nn.Bilinear):
            torch.nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0.0)

    def forward(self, c, h_pl, h_mi, s_bias1=None, s_bias2=None):
        c_x = torch.unsqueeze(c, 0)
        c_x = c_x.expand_as(h_pl)
        sc_1 = torch.squeeze(self.f_k(h_pl, c_x), 1)
        sc_2 = torch.squeeze(self.f_k(h_mi, c_x), 1)
        if s_bias1 is not None:
            sc_1 += s_bias1
        if s_bias2 is not None:
            sc_2 += s_bias2
        logits = torch.cat((sc_1, sc_2), 0)
        # print("testing, shape of logits: ", logits.size())
        return logits


class DGI(nn.Module):
    def __init__(self, in_dim, hidden_dim, out_dim, num_heads, use_residual=False):
        super(DGI, self).__init__()
        self.gat = GAT(in_dim, hidden_dim, out_dim, num_heads, use_residual)
        self.read = AvgReadout()
        self.sigm = nn.Sigmoid()
        self.disc = Discriminator(out_dim)

    def forward(self, nf):
        h_1 = self.gat(nf, False)
        c = self.read(h_1)
        c = self.sigm(c)
        h_2 = self.gat(nf, True)
        ret = self.disc(c, h_1, h_2)
        return h_1, ret

    # Detach the return variables
    def embed(self, nf):
        h_1 = self.gat(nf, False)
        return h_1.detach()


class OnlineTripletLoss(nn.Module):
    """
    Online Triplets loss
    Takes a batch of embeddings and corresponding labels.
    Triplets are generated using triplet_selector object that take embeddings and targets and return indices of
    triplets
    """

    def __init__(self, margin, triplet_selector):
        super(OnlineTripletLoss, self).__init__()
        self.margin = margin
        self.triplet_selector = triplet_selector

    def forward(self, embeddings, target):
        triplets = self.triplet_selector.get_triplets(embeddings, target)

        if embeddings.is_cuda:
            triplets = triplets.cuda()

        ap_distances = (embeddings[triplets[:, 0]] - embeddings[triplets[:, 1]]).pow(2).sum(1)  # .pow(.5)
        an_distances = (embeddings[triplets[:, 0]] - embeddings[triplets[:, 2]]).pow(2).sum(1)  # .pow(.5)
        losses = F.relu(ap_distances - an_distances + self.margin)

        return losses.mean(), len(triplets)


def pdist(vectors):
    distance_matrix = -2 * vectors.mm(torch.t(vectors)) + vectors.pow(2).sum(dim=1).view(1, -1) + vectors.pow(2).sum(
        dim=1).view(-1, 1)
    return distance_matrix


class TripletSelector:
    """
    Implementation should return indices of anchors, positive and negative samples
    return np array of shape [N_triplets x 3]
    """

    def __init__(self):
        pass

    def get_triplets(self, embeddings, labels):
        raise NotImplementedError


class FunctionNegativeTripletSelector(TripletSelector):
    """
    For each positive pair, takes the hardest negative sample (with the greatest triplet loss value) to create a triplet
    Margin should match the margin used in triplet loss.
    negative_selection_fn should take array of loss_values for a given anchor-positive pair and all negative samples
    and return a negative index for that pair
    """

    def __init__(self, margin, negative_selection_fn, cpu=True):
        super(FunctionNegativeTripletSelector, self).__init__()
        self.cpu = cpu
        self.margin = margin
        self.negative_selection_fn = negative_selection_fn

    def get_triplets(self, embeddings, labels):
        if self.cpu:
            embeddings = embeddings.cpu()
        distance_matrix = pdist(embeddings)
        distance_matrix = distance_matrix.cpu()

        labels = labels.cpu().data.numpy()
        triplets = []

        for label in set(labels):
            label_mask = (labels == label)
            label_indices = np.where(label_mask)[0]
            if len(label_indices) < 2:
                continue
            negative_indices = np.where(np.logical_not(label_mask))[0]
            anchor_positives = list(combinations(label_indices, 2))  # All anchor-positive pairs
            anchor_positives = np.array(anchor_positives)

            ap_distances = distance_matrix[anchor_positives[:, 0], anchor_positives[:, 1]]
            for anchor_positive, ap_distance in zip(anchor_positives, ap_distances):
                loss_values = ap_distance - distance_matrix[
                    torch.LongTensor(np.array([anchor_positive[0]])), torch.LongTensor(negative_indices)] + self.margin
                loss_values = loss_values.data.cpu().numpy()
                hard_negative = self.negative_selection_fn(loss_values)
                if hard_negative is not None:
                    hard_negative = negative_indices[hard_negative]
                    triplets.append([anchor_positive[0], anchor_positive[1], hard_negative])

        if len(triplets) == 0:
            triplets.append([anchor_positive[0], anchor_positive[1], negative_indices[0]])

        triplets = np.array(triplets)

        return torch.LongTensor(triplets)


def random_hard_negative(loss_values):
    hard_negatives = np.where(loss_values > 0)[0]
    return np.random.choice(hard_negatives) if len(hard_negatives) > 0 else None


def hardest_negative(loss_values):
    hard_negative = np.argmax(loss_values)
    return hard_negative if loss_values[hard_negative] > 0 else None


def HardestNegativeTripletSelector(margin, cpu=False): return FunctionNegativeTripletSelector(margin=margin,
                                                                                              negative_selection_fn=hardest_negative,
                                                                                              cpu=cpu)


def RandomNegativeTripletSelector(margin, cpu=False): return FunctionNegativeTripletSelector(margin=margin,
                                                                                             negative_selection_fn=random_hard_negative,
                                                                                             cpu=cpu)


# utils
# Dataset
class SocialDataset(Dataset):
    def __init__(self, path, index):
        self.features = np.load(path + '/' + str(index) + '/features.npy')
        temp = np.load(path + '/' + str(index) + '/labels.npy', allow_pickle=True)
        self.labels = np.asarray([int(each) for each in temp])
        self.matrix = self.load_adj_matrix(path, index)

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):
        return self.features[idx], self.labels[idx]

    def load_adj_matrix(self, path, index):
        s_bool_A_tid_tid = sparse.load_npz(path + '/' + str(index) + '/s_bool_A_tid_tid.npz')
        print("Sparse binary adjacency matrix loaded.")
        return s_bool_A_tid_tid

    # Used by remove_obsolete mode 1
    def remove_obsolete_nodes(self, indices_to_remove=None):  # indices_to_remove: list
        # torch.range(0, (self.labels.shape[0] - 1), dtype=torch.long)
        if indices_to_remove is not None:
            all_indices = np.arange(0, self.labels.shape[0]).tolist()
            indices_to_keep = list(set(all_indices) - set(indices_to_remove))
            self.features = self.features[indices_to_keep, :]
            self.labels = self.labels[indices_to_keep]
            self.matrix = self.matrix[indices_to_keep, :]  # keep row
            self.matrix = self.matrix[:, indices_to_keep]  # keep column
            #  remove nodes from matrix


if __name__ == '__main__':
    # from data_sets import Event2012_Dataset, Event2018_Dataset, MAVEN_Dataset, Arabic_Dataset

    args = args_define().args
    dataset = DatasetLoader("arabic_twitter").load_data()
    kpgnn = KPGNN(args, dataset)

    kpgnn.preprocess()
    kpgnn.fit()
    predictions, ground_truths = kpgnn.detection()

    # Evaluate model
    kpgnn.evaluate(predictions, ground_truths)


<<<<<<< HEAD
from . import detector
from . import dataset
from . import metrics
from . import utils


__all__ = ['detector', 'dataset', 'metrics', 'utils']
=======
# -*- coding: utf-8 -*-


from . import dataset
from . import metrics
from . import utils
from . import detector

__all__ = ['dataset', 'metrics', 'utils', 'detector']
>>>>>>> 52773300149e147b47eace9803e3651c4f43f810



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='SocialED',
    version='1.1.0',
    packages=find_packages(),
    author='beici',
    author_email='zhangkun23@buaa.edu.cn',
    description='A Python Library for Social Event Detection',
    install_requires=requirements,
    include_package_data=True,
    long_description=open('PyPI.rst', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/RingBDStack/SocialED',
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)



import os
import pytest
import pandas as pd
from unittest.mock import patch

import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from detector.LDA import LDA
from dataset.dataloader import DatasetLoader

from gensim.models.ldamodel import LdaModel
from gensim import corpora

# 设置测试数据路径
TEST_DATA_PATH = os.path.join(os.path.dirname(__file__), 'test_data')


# 创建一个测试数据集
@pytest.fixture
def test_dataset():
    data = {
        'filtered_words': [['word1', 'word2'], ['word3', 'word4'], ['word5', 'word6']],
        'event_id': [1, 2, 3]
    }
    return pd.DataFrame(data)


# 测试预处理方法
def test_preprocess(test_dataset):
    lda = LDA(dataset=test_dataset)
    df = lda.preprocess()
    assert 'processed_text' in df.columns
    assert df['processed_text'].tolist() == [['word1', 'word2'], ['word3', 'word4'], ['word5', 'word6']]


# 测试创建语料库方法
def test_create_corpus(test_dataset):
    lda = LDA(dataset=test_dataset)
    lda.preprocess()
    corpus, dictionary = lda.create_corpus(lda.df, 'processed_text')
    assert isinstance(corpus, list)
    assert isinstance(dictionary, corpora.Dictionary)


# 测试模型训练方法
def test_fit(test_dataset, tmpdir):
    lda = LDA(dataset=test_dataset, file_path=str(tmpdir))
    lda.preprocess()
    lda.fit()
    assert os.path.exists(os.path.join(tmpdir, 'lda_model'))


# 测试模型加载方法
def test_load_model(test_dataset, tmpdir):
    lda = LDA(dataset=test_dataset, file_path=str(tmpdir))
    lda.preprocess()
    lda.fit()
    loaded_model = lda.load_model()
    assert isinstance(loaded_model, LdaModel)


# 测试主题显示方法
def test_display_topics(test_dataset, tmpdir, capsys):
    lda = LDA(dataset=test_dataset, file_path=str(tmpdir))
    lda.preprocess()
    lda.fit()
    lda.load_model()
    lda.display_topics()
    captured = capsys.readouterr()
    assert "Topic" in captured.out


# 测试检测方法
def test_detection(test_dataset, tmpdir):
    lda = LDA(dataset=test_dataset, file_path=str(tmpdir))
    lda.preprocess()
    lda.fit()
    ground_truths, predictions = lda.detection()
    assert isinstance(ground_truths, list)
    assert isinstance(predictions, list)
    assert len(ground_truths) == len(predictions)


# 测试评估方法
def test_evaluate(test_dataset, tmpdir):
    lda = LDA(dataset=test_dataset, file_path=str(tmpdir))
    lda.preprocess()
    lda.fit()
    ground_truths, predictions = lda.detection()
    lda.evaluate(ground_truths, predictions)
    assert os.path.exists(os.path.join(tmpdir, 'lda_model_evaluation.txt'))


# 测试主函数
@patch('dataset.dataloader.DatasetLoader.load_data')
def test_main(mock_load_data, tmpdir):
    # 模拟数据集加载
    mock_load_data.return_value = pd.DataFrame({
        'filtered_words': [['word1', 'word2'], ['word3', 'word4'], ['word5', 'word6']],
        'event_id': [1, 2, 3]
    })

    lda = LDA(dataset=DatasetLoader("event2012").load_data(), file_path=str(tmpdir))
    lda.preprocess()
    lda.fit()
    ground_truths, predictions = lda.detection()
    lda.evaluate(ground_truths, predictions)
    assert os.path.exists(os.path.join(tmpdir, 'lda_model'))
    assert os.path.exists(os.path.join(tmpdir, 'unique_ground_truths_predictions.csv'))
    assert os.path.exists(os.path.join(tmpdir, 'lda_model_evaluation.txt'))



import argparse
import os
import pandas as pd
import numpy as np
from gensim.models import Word2Vec
from sklearn.model_selection import train_test_split
from sklearn import metrics
import logging
from sklearn.cluster import KMeans
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from dataset.dataloader import DatasetLoader
# Setup logging
# logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.INFO)


class WORD2VEC:
    def __init__(self,
                 dataset=DatasetLoader("arabic_twitter").load_data(),
                 vector_size=100, 
                 window=5,
                 min_count=1,
                 sg=1,
                 file_path='../model/model_saved/Word2vec/word2vec_model.model'):
        # print("ASDASd")
        # exit()
        self.dataset = dataset
        self.vector_size = vector_size
        self.window = window
        self.min_count = min_count
        self.sg = sg
        self.file_path = file_path
        self.df = None
        self.train_df = None
        self.test_df = None
        self.word2vec_model = None

    def preprocess(self):
        """
        Data preprocessing: tokenization, stop words removal, etc.
        """
        df = self.dataset[['filtered_words', 'event_id']].copy()
        df['processed_text'] = df['filtered_words'].apply(
            lambda x: [str(word).lower() for word in x] if isinstance(x, list) else [])
        self.df = df
        return df

    def fit(self):
        """
        Train the Word2Vec model and save it to a file.
        """
        # Ensure the directory exists
        os.makedirs(os.path.dirname(self.file_path), exist_ok=True)

        train_df, test_df = train_test_split(self.df, test_size=0.2, random_state=42)
        self.train_df = train_df
        self.test_df = test_df

        sentences = train_df['processed_text'].tolist()

        logging.info("Training Word2Vec model...")
        word2vec_model = Word2Vec(sentences=sentences, vector_size=self.vector_size, window=self.window,
                                  min_count=self.min_count, sg=self.sg)
        logging.info("Word2Vec model trained successfully.")

        # Save the trained model to a file
        word2vec_model.save(self.file_path)
        logging.info(f"Word2Vec model saved to {self.file_path}")

        self.word2vec_model = word2vec_model
        return word2vec_model

    def load_model(self):
        """
        Load the Word2Vec model from a file.
        """
        logging.info(f"Loading Word2Vec model from {self.file_path}...")
        word2vec_model = Word2Vec.load(self.file_path)
        logging.info("Word2Vec model loaded successfully.")

        self.word2vec_model = word2vec_model
        return word2vec_model

    def document_vector(self, document):
        """
        Create a document vector by averaging the Word2Vec embeddings of its words.
        """
        words = [word for word in document if word in self.word2vec_model.wv]
        if words:
            return np.mean(self.word2vec_model.wv[words], axis=0)
        else:
            return np.zeros(self.vector_size)

    def detection(self):
        """
        Detect events by representing each document as the average Word2Vec embedding of its words.
        """
        self.load_model()  # Ensure the model is loaded before making detections

        test_vectors = self.test_df['processed_text'].apply(self.document_vector)
        predictions = np.stack(test_vectors)

        ground_truths = self.test_df['event_id'].tolist()
        kmeans = KMeans(n_clusters=len(set(ground_truths)), random_state=42)
        predictions = kmeans.fit_predict(predictions)

        return ground_truths, predictions

    def evaluate(self, ground_truths, predictions):
        """
        Evaluate the model.
        """

        # Calculate Adjusted Rand Index (ARI)
        ari = metrics.adjusted_rand_score(ground_truths, predictions)
        print(f"Adjusted Rand Index (ARI): {ari}")

        # Calculate Adjusted Mutual Information (AMI)
        ami = metrics.adjusted_mutual_info_score(ground_truths, predictions)
        print(f"Adjusted Mutual Information (AMI): {ami}")

        # Calculate Normalized Mutual Information (NMI)
        nmi = metrics.normalized_mutual_info_score(ground_truths, predictions)
        print(f"Normalized Mutual Information (NMI): {nmi}")

        return ari, ami, nmi


# Main function
if __name__ == "__main__":
    # from data_sets import Event2012_Dataset, Event2018_Dataset, MAVEN_Dataset, Arabic_Dataset


    word2vec = WORD2VEC()

    # Data preprocessing
    word2vec.preprocess()

    # Train the model
    word2vec.fit()

    # detection
    ground_truths, predictions = word2vec.detection()

    # Evaluate the model
    word2vec.evaluate(ground_truths, predictions)



import argparse
import os
import logging
import datetime
import pandas as pd
import numpy as np
from gensim.models.ldamodel import LdaModel
from gensim import corpora
from sklearn.model_selection import train_test_split
from sklearn import metrics
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from dataset.dataloader import DatasetLoader

class LDA:
    def __init__(self,
                 dataset=DatasetLoader("arabic_twitter").load_data(),
                 num_topics=50,
                 passes=20,
                 iterations=50,
                 alpha='symmetric',
                 eta=None,
                 random_state=1,
                 eval_every=10,
                 chunksize=2000,
                 file_path='../model/model_saved/LDA/'):
        self.dataset = dataset
        self.num_topics = num_topics
        self.passes = passes
        self.iterations = iterations
        self.alpha = alpha
        self.eta = eta
        self.random_state = random_state
        self.eval_every = eval_every
        self.chunksize = chunksize
        self.df = None
        self.train_df = None
        self.test_df = None
        self.file_path = file_path
        self.model_path = os.path.join(file_path, 'lda_model')

    def preprocess(self):
        """
        Data preprocessing: tokenization, stop words removal, etc.
        """
        df = self.dataset[['filtered_words', 'event_id']].copy()
        df['processed_text'] = df['filtered_words'].apply(
            lambda x: [str(word).lower() for word in x] if isinstance(x, list) else [])
        self.df = df
        return df

    def create_corpus(self, df, text_column):
        """
        Create corpus and dictionary required for LDA model.
        """
        texts = df[text_column].tolist()
        dictionary = corpora.Dictionary(texts)
        corpus = [dictionary.doc2bow(text) for text in texts]
        return corpus, dictionary

    def load_model(self):
        """
        Load the LDA model from a file.
        """
        logging.info(f"Loading LDA model from {self.model_path}...")
        lda_model = LdaModel.load(self.model_path)
        logging.info("LDA model loaded successfully.")

        self.lda_model = lda_model
        return lda_model

    def display_topics(self, num_words=10):
        """
        Display topics generated by the LDA model.
        """
        topics = self.lda_model.show_topics(num_words=num_words, formatted=False)
        for i, topic in topics:
            print(f"Topic {i}: {[word for word, _ in topic]}")

    def fit(self):
        os.makedirs(os.path.dirname(self.model_path), exist_ok=True)

        train_df, test_df = train_test_split(self.df, test_size=0.2, random_state=self.random_state)
        self.train_df = train_df
        self.test_df = test_df
        train_corpus, train_dictionary = self.create_corpus(train_df, 'processed_text')

        logging.info("Training LDA model...")
        lda_model = LdaModel(corpus=train_corpus, id2word=train_dictionary, num_topics=self.num_topics,
                             passes=self.passes,
                             iterations=self.iterations, alpha=self.alpha, eta=self.eta, random_state=self.random_state,
                             eval_every=self.eval_every, chunksize=self.chunksize)
        logging.info("LDA model trained successfully.")

        # Save the trained model to a file
        lda_model.save(self.model_path)
        logging.info(f"LDA model saved to {self.model_path}")

    def detection(self):
        """
        Assign topics to each document and save unique ground truths and predictions to a CSV file.
        """
        self.load_model()  # Ensure the model is loaded before making detections
        corpus, _ = self.create_corpus(self.test_df, 'processed_text')
        topics = [self.lda_model.get_document_topics(bow) for bow in corpus]

        # Get the ground truth labels and predicted labels
        ground_truths = self.test_df['event_id'].tolist()
        predictions = [max(topic, key=lambda x: x[1])[0] for topic in topics]

        # Convert to sets to remove duplicates
        unique_ground_truths = list(set(ground_truths))
        unique_predictions = list(set(predictions))

        # Pad the shorter list with None to make them the same length
        max_len = max(len(unique_ground_truths), len(unique_predictions))
        unique_ground_truths.extend([None] * (max_len - len(unique_ground_truths)))
        unique_predictions.extend([None] * (max_len - len(unique_predictions)))

        # Combine into a dataframe
        data = {
            'Unique Ground Truths': unique_ground_truths,
            'Unique Predictions': unique_predictions
        }
        df = pd.DataFrame(data)

        # Save to a CSV file
        output_file = os.path.join(self.file_path, "unique_ground_truths_predictions.csv")
        df.to_csv(output_file, index=False)
        print(f"Unique ground truths and predictions have been saved to {output_file}")

        return ground_truths, predictions

    def evaluate(self, ground_truths, predictions):
        """
        Evaluate the model.
        """
        # Calculate Normalized Mutual Information (NMI)
        nmi = metrics.normalized_mutual_info_score(ground_truths, predictions)
        print(f"Normalized Mutual Information (NMI): {nmi}")

        # Calculate Adjusted Mutual Information (AMI)
        ami = metrics.adjusted_mutual_info_score(ground_truths, predictions)
        print(f"Adjusted Mutual Information (AMI): {ami}")

        # Calculate Adjusted Rand Index (ARI)
        ari = metrics.adjusted_rand_score(ground_truths, predictions)
        print(f"Adjusted Rand Index (ARI): {ari}")

        # Get the current date and time
        current_datetime = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # Save results to a file in append mode
        with open(self.model_path + "_evaluation.txt", "a") as f:
            f.write(f"Date and Time: {current_datetime}\n")
            f.write(f"Normalized Mutual Information (NMI): {nmi}\n")
            f.write(f"Adjusted Mutual Information (AMI): {ami}\n")
            f.write(f"Adjusted Rand Index (ARI): {ari}\n")
            f.write("\n")  # Add a newline for better readability


# Main function
if __name__ == "__main__":
    dataset = DatasetLoader("maven").load_data()

    lda = LDA(dataset)

    # Data preprocessing
    lda.preprocess()

    # Train the LDA model
    lda.fit()

    # detection
    ground_truths, predictions = lda.detection()

    # Evaluate the model
    lda.evaluate(ground_truths, predictions)



#!/usr/bin/env python
# encoding: utf-8
import logging
import argparse
import gc
import torch.optim as optim
from ignite.contrib.handlers import ProgressBar
from ignite.engine import Engine, Events
from ignite.handlers import EarlyStopping, ModelCheckpoint
from ignite.metrics import RunningAverage, Average
from sklearn.metrics import accuracy_score
from torch.optim.lr_scheduler import StepLR
from transformers import AutoTokenizer, AutoModel, AutoConfig
from collections import namedtuple, OrderedDict, Counter
from typing import Any, List
import math
import os
import random
import numpy as np
import torch
from torch import nn
from sklearn import metrics, manifold
from sklearn.cluster import KMeans, DBSCAN, OPTICS
from hdbscan import HDBSCAN
from matplotlib import pyplot as plt
from concurrent.futures import ThreadPoolExecutor, as_completed
import datetime
import itertools
import scipy as sp
from sklearn.model_selection import train_test_split
import sys

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from dataset.dataloader import DatasetLoader

COLUMNS_12 = ["event_id", "tweet_id", "text", "user_id", "created_at", "user_loc", "place_type",
              "place_full_name", "place_country_code", "hashtags", "user_mentions", "urls",
              "entities", "words", "filtered_words", "sampled_words"]
DataItem = namedtuple('DataItem', COLUMNS_12)

logging.basicConfig(level=logging.WARN,
                    format="%(asctime)s %(name)s %(levelname)s %(message)s",
                    datefmt='%Y-%m-%d  %H:%M:%S %a')


class args_define:
    def __init__(self, **kwargs):
        # Hyper parameters
        self.dataset = kwargs.get('dataset', '../model/model_saved/rplmsed/cache/twitter12.npy')
        self.plm_path = kwargs.get('plm_path', '../model/model_needed/base_plm_model/roberta-large')
        self.file_path = kwargs.get('file_path', '../model/model_saved/rplmsed/')
        self.plm_tuning = kwargs.get('plm_tuning', False)
        self.use_ctx_att = kwargs.get('use_ctx_att', False)
        self.offline = kwargs.get('offline', True)
        self.ctx_att_head_num = kwargs.get('ctx_att_head_num', 2)
        self.pmt_feats = kwargs.get('pmt_feats', (0, 1, 2, 4))
        self.batch_size = kwargs.get('batch_size', 128)
        # self.batch_size = kwargs.get('batch_size', 32)
        self.lmda1 = kwargs.get('lmda1', 0.010)
        self.lmda2 = kwargs.get('lmda2', 0.005)
        self.tao = kwargs.get('tao', 0.90)
        self.optimizer = kwargs.get('optimizer', 'Adam')
        self.lr = kwargs.get('lr', 2e-5)
        self.weight_decay = kwargs.get('weight_decay', 1e-5)
        self.momentum = kwargs.get('momentum', 0.9)
        self.step_lr_gamma = kwargs.get('step_lr_gamma', 0.98)
        self.max_epochs = kwargs.get('max_epochs', 1)
        self.ckpt_path = kwargs.get('ckpt_path', '../model/model_saved/rplmsed/ckpt/')
        self.eva_data = kwargs.get('eva_data', "../model/model_saved/rplmsed/Eva_data/")
        self.early_stop_patience = kwargs.get('early_stop_patience', 2)
        self.early_stop_monitor = kwargs.get('early_stop_monitor', 'loss')
        self.SAMPLE_NUM_TWEET = kwargs.get('SAMPLE_NUM_TWEET', 60)
        self.WINDOW_SIZE = kwargs.get('WINDOW_SIZE', 3)
        self.device = kwargs.get('device', "cuda:0" if torch.cuda.is_available() else "cpu")

        # Store all arguments in a single attribute
        self.args = argparse.Namespace(**{
            'dataset': self.dataset,
            'plm_path': self.plm_path,
            'file_path': self.file_path,
            'plm_tuning': self.plm_tuning,
            'use_ctx_att': self.use_ctx_att,
            'offline': self.offline,
            'ctx_att_head_num': self.ctx_att_head_num,
            'pmt_feats': self.pmt_feats,
            'batch_size': self.batch_size,
            'lmda1': self.lmda1,
            'lmda2': self.lmda2,
            'tao': self.tao,
            'optimizer': self.optimizer,
            'lr': self.lr,
            'weight_decay': self.weight_decay,
            'momentum': self.momentum,
            'step_lr_gamma': self.step_lr_gamma,
            'max_epochs': self.max_epochs,
            'ckpt_path': self.ckpt_path,
            'eva_data': self.eva_data,
            'early_stop_patience': self.early_stop_patience,
            'early_stop_monitor': self.early_stop_monitor,
            'SAMPLE_NUM_TWEET': self.SAMPLE_NUM_TWEET,
            'WINDOW_SIZE': self.WINDOW_SIZE,
            'device': self.device
        })


# class args_define():
#     def list_of_ints(arg):
#         return list(map(int, arg.split(',')))
#
#     parser = argparse.ArgumentParser()
#     parser.add_argument('--dataset', type=str, default='../model/model_saved/rplmsed/cache/twitter12.npy')
#     parser.add_argument('--plm_path', type=str, default='../model/model_needed/base_plm_model/roberta-large')
#     parser.add_argument('--file_path', type=str, default='../model/model_saved/rplmsed/')
#     parser.add_argument('--plm_tuning', action='store_true')
#     parser.add_argument('--use_ctx_att', action='store_true')
#     # parser.add_argument('--offline', action='store_true')
#     parser.add_argument('--offline', type=bool, default=True)
#     parser.add_argument('--ctx_att_head_num', type=int, default=2)
#     parser.add_argument("--pmt_feats", type=list_of_ints, default=(0, 1, 2, 4),
#                         help="(entities, hashtags, user, words, time)")
#
#     parser.add_argument("--batch_size", type=int, default=128, help="Batch size for training")
#     parser.add_argument("--lmda1", type=float, default=0.010)
#     parser.add_argument("--lmda2", type=float, default=0.005)
#     parser.add_argument("--tao", type=float, default=0.90)
#
#     parser.add_argument("--optimizer", type=str, default='Adam', help="Optimizer, Adam, AdamW or SGD")
#     parser.add_argument("--lr", type=float, default=2e-5, help="Learning rate")
#     parser.add_argument("--weight_decay", type=float, default=1e-5, help="weight decay")
#     parser.add_argument("--momentum", type=float, default=0.9, help="momentum for SGD optimizer")
#     parser.add_argument("--step_lr_gamma", type=float, default=0.98, help="gamma for step learning rate schedule")
#
#     parser.add_argument("--max_epochs", type=int, default=1, help="Number of training epochs")
#     # ablation_ckpt  ablation_Eva_datas
#     parser.add_argument('--ckpt_path', type=str, default='../model/model_saved/rplmsed/ckpt/',
#                         help='path to checkpoint files')
#     parser.add_argument("--eva_data", type=str, default="../model/model_saved/rplmsed/Eva_data/",
#                         help="path to Evaluate_datas")
#
#     parser.add_argument("--early_stop_patience", type=int, default=2)
#     parser.add_argument("--early_stop_monitor", type=str, default='loss', help='loss')
#
#     parser.add_argument("--SAMPLE_NUM_TWEET", type=int, default=60)
#     parser.add_argument("--WINDOW_SIZE", type=int, default=3)
#
#     parser.add_argument("--device", type=str,
#                         default="cuda:0" if torch.cuda.is_available() else "cpu",
#                         help="Device (cuda or cpu)")
#
#     args = parser.parse_args()


class Preprocessor:
    args = args_define().args

    def __init__(self):
        pass

    def twitter12_process(self, dataset):
        os.makedirs('../model/model_saved/rplmsed/cache', exist_ok=True)
        print(f"load data  ... ")
        df = dataset
        np_data = df.to_numpy()
        print("\tDone")

        blk_data = self.pre_process(np_data)
        print(f"save data to '{self.dataset}' ... ", end='')
        np.save(f'{self.dataset}', blk_data)
        print("\tDone")

    def to_sparse_matrix(self, feat_to_tw, tw_num, tao=0):
        tw_adj = sp.sparse.coo_matrix((tw_num, tw_num), dtype=np.int8)
        tw_adj = tw_adj.todok()  # convert to dok
        for f in feat_to_tw.keys():
            for i in feat_to_tw[f]:
                for j in feat_to_tw[f]:
                    tw_adj[i, j] += 1

        tw_adj = tw_adj > tao
        tw_adj = tw_adj.tocsr().astype(np.int8)
        return tw_adj

    def build_entity_adj(self, data):
        tw_num = len(data)
        feat_to_tw = {}
        for i, it in enumerate(data):
            feats = it.entities
            feats = [e for e, t in feats]

            for f in feats:
                f = f.lower()
                if f not in feat_to_tw:
                    feat_to_tw[f] = set()
                feat_to_tw[f].add(i)

        return self.to_sparse_matrix(feat_to_tw, tw_num)

    def build_hashtag_adj(self, data):
        tw_num = len(data)
        feat_to_tw = {}
        for i, it in enumerate(data):
            feats = it.hashtags

            for f in feats:
                f = f.lower()
                if f not in feat_to_tw:
                    feat_to_tw[f] = set()
                feat_to_tw[f].add(i)

        return self.to_sparse_matrix(feat_to_tw, tw_num)

    def build_words_adj(self, data):
        tw_num = len(data)
        feat_to_tw = {}
        for i, it in enumerate(data):
            feats = it.words

            for f in feats:
                f = f.lower()
                if f not in feat_to_tw:
                    feat_to_tw[f] = set()
                feat_to_tw[f].add(i)

        return self.to_sparse_matrix(feat_to_tw, tw_num)

    def build_user_adj(self, data):
        tw_num = len(data)
        feat_to_tw = {}
        for i, it in enumerate(data):
            feats = it.user_mentions
            feats.append(it.user_id)

            for f in feats:
                if f not in feat_to_tw:
                    feat_to_tw[f] = set()
                feat_to_tw[f].add(i)

        return self.to_sparse_matrix(feat_to_tw, tw_num)

    def build_creat_at_adj(self, data):
        tw_num = len(data)
        tw_feat_idx = []
        feat_to_idx = {}
        for i, it in enumerate(data):
            feats = it.created_at
            feats = [e for e, t in feats]

            for f in feats:
                if f not in feat_to_idx:
                    feat_to_idx[f] = len(feat_to_idx)
                f_idx = feat_to_idx[f]

                tw_feat_idx.append([i, f_idx])

        tw_feat_val = np.ones((len(tw_feat_idx),), dtype=np.int32)
        tw_feat_idx = np.array(tw_feat_idx, dtype=np.int64).T

        feat_num = len(feat_to_idx)
        tw_feat_mat = sp.sparse.coo_matrix(
            (tw_feat_val, (tw_feat_idx[0, :], tw_feat_idx[1, :])),
            shape=(tw_num, feat_num),
            dtype=np.int8)

        tw_adj = tw_feat_mat @ tw_feat_mat.T
        return tw_adj

    def tweet_to_event(self, data):
        ev_ids = sorted(set(it.event_id for it in data))
        ev_to_idx = {eid: i for i, eid in enumerate(ev_ids)}
        tw_to_ev = [ev_to_idx[it.event_id] for it in data]
        return tw_to_ev, ev_to_idx

    def build_feats_adj(self, data, feats):
        feats_adj = [func(data) for f, func in feats]
        return feats_adj

    def build_feat_adj(self, data, cols):
        tw_num = len(data)
        tw_feat_idx = []
        feat_to_idx = {}
        cols = [DataItem._fields.index(c) for c in cols] if isinstance(cols, list) else [DataItem._fields.index(cols)]
        for i, it in enumerate(data):
            feats = [
                list(itertools.chain(*it[c])) if isinstance(it[c], list) or isinstance(it[c], tuple) else [it[c]]
                for c in cols  # 特征列
            ]
            feats = [f for cf in feats for f in cf]

            for f in feats:
                if f not in feat_to_idx:
                    feat_to_idx[f] = len(feat_to_idx)
                f_idx = feat_to_idx[f]

                tw_feat_idx.append([i, f_idx])

        tw_feat_val = np.ones((len(tw_feat_idx),), dtype=np.int32)
        tw_feat_idx = np.array(tw_feat_idx, dtype=np.int64).T

        feat_num = len(feat_to_idx)
        tw_feat_mat = sp.sparse.coo_matrix(
            (tw_feat_val, (tw_feat_idx[0, :], tw_feat_idx[1, :])),
            shape=(tw_num, feat_num),
            dtype=np.int8)

        tw_adj = tw_feat_mat @ tw_feat_mat.T
        return tw_adj

    def get_time_relation(self, tw_i, tw_j, delta: datetime.timedelta = datetime.timedelta(hours=4)):
        a, b = tw_i.created_at, tw_j.created_at
        return int(a - b < delta) if a > b else int(b - a < delta)

    def make_train_samples(self, tw_adj, tw_to_ev, data):
        tw_adj_num = len(tw_adj)
        tw_num = len(tw_to_ev)
        ev_num = max(tw_to_ev) + 1

        tw_ev_mat = np.zeros(shape=(tw_num, ev_num), dtype=np.int8)
        for i, e in enumerate(tw_to_ev):
            tw_ev_mat[i, e] = 1

        eye = sp.sparse.eye(tw_num, tw_num, dtype=np.int8)
        adj = tw_adj[0] - eye
        for f in range(1, tw_adj_num):
            adj = adj + (tw_adj[f] - eye)

        adj = np.asarray(adj.todense())

        pairs = []
        for i in range(tw_num):
            ev_idx = tw_to_ev[i]
            ev_tw_vec = tw_ev_mat[:, ev_idx]
            ev_tw_num = ev_tw_vec.sum()
            if ev_tw_num < 5:
                # print(f"outlier or small events: {i} -- {tw_to_ev[i]}--{ev_tw_num[tw_to_ev[i]]}")
                continue

            adj_i_tw = adj[i, :]
            adj_i_tw_score = np.exp(adj_i_tw - (1. - ev_tw_vec) * 1e12)

            pos_idx, = np.nonzero(ev_tw_vec)
            p = sp.special.softmax(adj_i_tw_score.take(pos_idx))

            pos_idx = np.random.choice(pos_idx, size=args.SAMPLE_NUM_TWEET, p=p)
            # (tag, event, (tweet_a, tweet_b), [feats,])
            pos_pairs = [
                (
                    int(tw_to_ev[i] == tw_to_ev[j]), tw_to_ev[i], (i, j),
                    list(1 if tw_adj[f][i, j] > 0 else 0 for f in range(tw_adj_num)) + [
                        self.get_time_relation(data[i], data[j])]
                )
                for j in pos_idx
            ]
            pairs.extend(pos_pairs)

            neg_idx, = np.nonzero(1 - ev_tw_vec)
            adj_i_tw_score = np.exp(adj_i_tw - ev_tw_vec * 1e12)

            p = sp.special.softmax(adj_i_tw_score.take(neg_idx))

            neg_idx = np.random.choice(neg_idx, size=args.SAMPLE_NUM_TWEET, p=p)

            # (tag, event, (tweet_a, tweet_b), [feats,])
            neg_pairs = [
                (
                    int(tw_to_ev[i] == tw_to_ev[j]), tw_to_ev[i], (i, j),
                    list(1 if tw_adj[f][i, j] > 0 else 0 for f in range(tw_adj_num)) + [
                        self.get_time_relation(data[i], data[j])]
                )
                for j in neg_idx
            ]
            pairs.extend(neg_pairs)

        return pairs

    def make_ref_samples(self, tw_adj, tw_to_ev, data):
        tw_adj_num = len(tw_adj)
        tw_num = len(tw_to_ev)

        pairs = []
        adj = tw_adj[0]
        for f in range(1, tw_adj_num):
            adj = adj + tw_adj[f]

        adj = np.asarray(adj.todense())
        eye = np.eye(tw_num, tw_num, dtype=np.int8)
        adj = adj * (1 - eye) + eye

        tw_idx = np.arange(tw_num)
        for i in range(tw_num):
            p = sp.special.softmax(np.exp(adj[i]))

            ref_idx = np.random.choice(tw_idx, size=args.SAMPLE_NUM_TWEET * 3, p=p)
            # (tag, event, (tweet_a, tweet_b), [feats,])
            ref_pairs = [
                (
                    int(tw_to_ev[i] == tw_to_ev[j]),
                    tw_to_ev[i], (i, j),
                    list(1 if tw_adj[f][i, j] > 0 else 0 for f in range(tw_adj_num)) + [
                        self.get_time_relation(data[i], data[j])]
                )
                for j in ref_idx
            ]
            pairs.extend(ref_pairs)

        return pairs

    def process_block(self, block):
        blk = {}

        FEAT_COLS = [
            ("entities", self.build_entity_adj),
            ("hashtags", self.build_hashtag_adj),
            ("user", self.build_user_adj),  # user_mentions and user_id
            ("words", self.build_words_adj),

            # ("create_at", self.build_creat_at_adj)
        ]

        for name in ["train", "test", "valid"]:
            data = block[name]
            tw_to_ev, ev_to_idx = self.tweet_to_event(data)
            tw_adj = self.build_feats_adj(data, FEAT_COLS)

            blk[name] = {
                "data": data,
                "tw_to_ev": tw_to_ev,
                "ev_to_idx": ev_to_idx,
                "tw_adj": tw_adj
            }

            if name == "train" or name == "valid":
                if data:
                    blk[name]["samples"] = self.make_train_samples(tw_adj, tw_to_ev, data)
                else:
                    blk[name]["samples"] = []

            if name == "test":
                if data:
                    blk[name]["samples"] = self.make_ref_samples(tw_adj, tw_to_ev, data)
                else:
                    blk[name]["samples"] = []

        return blk

    def split_train_test_validation(self, data: List):
        block = []
        off_dataset = []
        for i in range(len(data)):
            if i == 0:
                data_size = len(data[i])
                valid_size = math.ceil(data_size * 0.2)
                test_size = math.ceil(data_size * 0.1)  # Add test size
                train, temp = train_test_split(data[i], test_size=valid_size + test_size, random_state=42, shuffle=True)
                valid, test = train_test_split(temp, test_size=test_size, random_state=42, shuffle=True)
                block.append({"train": train, "test": test, "valid": valid})

                print(f"Block {i}: Train size: {len(train)}, Valid size: {len(valid)}, Test size: {len(test)}")

                off_test_size = math.ceil(data_size * 0.2)
                off_valid_size = math.ceil(data_size * 0.1)
                off_train, off_test = train_test_split(data[i], test_size=off_test_size, random_state=42, shuffle=True)
                off_train, off_valid = train_test_split(off_train, test_size=off_valid_size, random_state=42,
                                                        shuffle=True)

                print("create offline dataset ...", end="\t")
                off_dataset.append(self.process_block({"train": off_train, "test": off_test, "valid": off_valid}))
                print("done")

                print(f"save data to '{args.file_path}cache/offline.npy' ... ", end='')
                np.save(args.file_path + 'cache/offline.npy', off_dataset)
                print("\tDone")

            elif i % args.WINDOW_SIZE == 0:
                sub_data = []
                for j in range(args.WINDOW_SIZE):
                    sub_data += data[i - j]
                sub_data_size = len(sub_data)
                sub_valid_size = math.ceil(sub_data_size * 0.2)
                train, valid = train_test_split(sub_data, test_size=sub_valid_size, random_state=42, shuffle=True)
                block.append({"train": train, "test": data[i], "valid": valid})
                print(f"Block {i}: Train size: {len(train)}, Valid size: {len(valid)}, Test size: {len(data[i])}")
            else:
                block.append({"train": [], "test": data[i], "valid": []})
                print(f"Block {i}: Train size: 0, Valid size: 0, Test size: {len(data[i])}")

        return block

    def split_into_blocks(self, data):

        data = [DataItem(*it) for it in data]
        data = sorted(data, key=lambda it: it.created_at)
        groups = itertools.groupby(data, key=lambda it: it.created_at.timetuple().tm_yday)
        groups = {k: list(g) for k, g in groups}

        days = sorted(groups.keys())
        blk0 = [groups[d] for d in days[:7]]
        blk0 = [it for b in blk0 for it in b]

        print(f"Initial Block 0: {len(blk0)} items")

        day_blk = [groups[d] for d in days[7:-1]]
        for idx, blk in enumerate(day_blk, start=1):
            print(f"Block {idx}: {len(blk)} items")

        blocks = [blk0] + day_blk
        datacount = [len(sublist) for sublist in blocks]

        print(f"save block datas counts into '{args.file_path}cache/datacount.npy' ", end='')
        os.makedirs(f'{args.file_path}cache', exist_ok=True)
        np.save(f'{args.file_path}cache/datacount.npy', datacount)
        print("done")

        return self.split_train_test_validation(blocks)

    def pre_process(self, data):
        print("split data into blocks... ")
        blocks = self.split_into_blocks(data)
        print("\tDone")

        print("process blocks..., ", end='')
        data_blocks = []
        for i, blk in enumerate(blocks):
            print(i, end=" ")
            blk = self.process_block(blk)
            data_blocks.append(blk)

        print("\tDone")
        return data_blocks


class RPLM_SED:
    def __init__(self, args, dataset):
        self.dataset = dataset

    def preprocess(self):
        preprocessor = Preprocessor()
        preprocessor.twitter12_process(self.dataset)

    def fit(self):
        torch.manual_seed(2357)
        args.model_name = os.path.basename(os.path.normpath(args.plm_path))
        dataset_name = os.path.basename(args.dataset)
        args.dataset_name = dataset_name.replace(".npy", "")

        if 'cuda' in args.device:
            torch.cuda.manual_seed(2357)

        tokenizer = AutoTokenizer.from_pretrained(args.plm_path)
        data_blocks = load_data_blocks(args.dataset, args, tokenizer)
        self.model = start_run(args, data_blocks)

    def detection(self):
        blk = torch.load(f'{args.file_path}cache/cache_long_tail/roberta-large-twitter12.npy')
        test = blk['test']

        msg_tags = np.array(test['tw_to_ev'], dtype=np.int32)
        tst_num = msg_tags.shape[0]
        msg_feats = torch.zeros((tst_num, self.model.feat_size()), device='cpu')
        ref_num = torch.zeros((tst_num,), dtype=torch.long, device='cpu')

        msg_feats = msg_feats / (ref_num + torch.eq(ref_num, 0).float()).unsqueeze(-1)
        msg_feats = msg_feats.numpy()

        n_clust = len(test['ev_to_idx'])
        k_means_score = run_kmeans(msg_feats, n_clust, msg_tags)

        k_means = KMeans(init="k-means++", n_clusters=n_clust, n_init=40, random_state=0)
        k_means.fit(msg_feats)

        predictions = k_means.labels_
        ground_truths = msg_tags
        return predictions, ground_truths

    def evaluate(self, predictions, ground_truths):
        ars = metrics.adjusted_rand_score(ground_truths, predictions)

        # Calculate Adjusted Mutual Information (AMI)
        ami = metrics.adjusted_mutual_info_score(ground_truths, predictions)

        # Calculate Normalized Mutual Information (NMI)
        nmi = metrics.normalized_mutual_info_score(ground_truths, predictions)

        print(f"Model Adjusted Rand Index (ARI): {ars}")
        print(f"Model Adjusted Mutual Information (AMI): {ami}")
        print(f"Model Normalized Mutual Information (NMI): {nmi}")
        return ars, ami, nmi


def get_model(args):
    return PairPfxTuningEncoder(
        len(args.pmt_feats), args.plm_path, args.plm_tuning,
        use_ctx_att=args.use_ctx_att, ctx_att_head_num=args.ctx_att_head_num)


def initialize(model, args, num_train_batch):
    # parameters = model.parameters()  # 优化器的初始化
    parameters = [
        {
            'name': 'pair_cls',
            'params': model.pair_cls.parameters(),
            'lr': args.lr
        }, {
            'name': 'pfx_embedding',
            'params': model.pfx_embedding.parameters(),
            'lr': args.lr
        }
    ]

    if args.plm_tuning:
        parameters.append(
            {
                'name': 'encoder',
                'params': model.plm.parameters(),
                'lr': args.lr / 100.
            }
        )

    if args.optimizer == 'Adam':
        optimizer = optim.Adam(parameters, lr=args.lr, weight_decay=args.weight_decay)
    elif args.optimizer == 'AdamW':
        optimizer = optim.AdamW(parameters, lr=args.lr, weight_decay=args.weight_decay)
    elif args.optimizer == 'RAdam':
        optimizer = optim.RAdam(parameters, lr=args.lr, weight_decay=args.weight_decay)
    elif args.optimizer == 'SGD':
        optimizer = optim.SGD(parameters, lr=args.lr, weight_decay=args.weight_decay, momentum=args.momentum)
    else:
        raise Exception("unsupported optimizer %s" % args.optimizer)

    lr_scheduler = None
    if args.step_lr_gamma > 0:
        lr_scheduler = StepLR(optimizer, step_size=num_train_batch, gamma=args.step_lr_gamma)

    return optimizer, lr_scheduler


def batch_to_tensor(batch, args):
    tags = [tag for tag, evt, a, b, fix, tok, _ in batch]
    events = [evt for tag, evt, a, b, fix, tok, _ in batch]
    prefix = [fix for tag, evt, a, b, fix, tok, _ in batch]
    toks = [tok for tag, evt, a, b, fix, tok, _ in batch]
    typs = [typ for tag, evt, a, b, fix, tok, typ in batch]

    max_len = min(max([len(it) for it in toks]), 200)
    toks = [pad_seq(it, pad=args.pad_tok_id, max_len=max_len) for it in toks]
    toks = torch.tensor(toks, dtype=torch.long)
    typs = [pad_seq(it, pad=args.pad_tok_id, max_len=max_len) for it in typs]
    typs = torch.tensor(typs, dtype=torch.long)
    tags = torch.tensor(tags, dtype=torch.long)
    events = torch.tensor(events, dtype=torch.long)
    prefix = torch.tensor(prefix, dtype=torch.long)

    return toks, typs, prefix, tags, events


# loss functions
# cls_loss = torch.nn.BCEWithLogitsLoss()

def create_trainer(model, optimizer, lr_scheduler, args):
    evt_proto = torch.zeros((args.train_evt_num, model.feat_size()), device=args.device, requires_grad=False)
    cls_loss = torch.nn.BCEWithLogitsLoss()

    # update event cluster center prototype by training batch
    def update_evt_proto(feats, events, alpha):
        proto = torch.zeros_like(evt_proto)
        proto.index_add_(0, events, feats)

        ev_idx, ev_idx_inv, ev_count = torch.unique(events, return_inverse=True, return_counts=True)
        proto_a = torch.index_select(proto, dim=0, index=ev_idx) / ev_count.unsqueeze(-1)
        proto_b = torch.index_select(evt_proto, dim=0, index=ev_idx)

        source = alpha * proto_a + (1 - alpha) * proto_b
        # source = proto_a
        source.detach_()
        source.requires_grad = False

        evt_proto.index_copy_(0, ev_idx, source)

        return proto_a

    # training logic for iteration
    def _train_step(_, batch):
        data = batch_to_tensor(batch, args)
        dist_loss = torch.nn.PairwiseDistance()

        toks, typs, prefix, tags, events = [x.to(args.device) for x in data]
        mask = torch.not_equal(toks, args.pad_tok_id).to(args.device)

        model.train()
        logit, left_feat = model(toks, typs, prefix, mask)

        loss = cls_loss(logit, tags.float())
        pred = torch.gt(logit, 0.)

        feats = left_feat
        evt_feats = update_evt_proto(feats, events, 0.8)
        protos = evt_proto.index_select(0, events)

        intra_dist = dist_loss(feats, protos)
        intra_dist_loss = intra_dist.mean()

        rand_idx = torch.randperm(evt_feats.size(0), device=args.device)
        rand_evt_feats = evt_feats.index_select(0, rand_idx)
        inter_dist_loss = torch.nn.functional.pairwise_distance(evt_feats, rand_evt_feats)

        inter_dist_loss = torch.maximum(10 - inter_dist_loss, torch.zeros_like(inter_dist_loss)).mean()

        if args.lmda1 > 0.:
            loss = loss + args.lmda1 * inter_dist_loss + args.lmda2 * intra_dist_loss
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        if lr_scheduler is not None:
            lr_scheduler.step()

        del toks, prefix, mask
        acc = accuracy_score(tags.cpu(), pred.cpu())

        return loss, acc, inter_dist_loss, intra_dist_loss

    # Define trainer engine
    trainer = Engine(_train_step)

    # metrics for trainer
    RunningAverage(output_transform=lambda x: x[0]).attach(trainer, 'loss')
    RunningAverage(output_transform=lambda x: x[1]).attach(trainer, 'acc')
    RunningAverage(output_transform=lambda x: x[2]).attach(trainer, 'inter')
    RunningAverage(output_transform=lambda x: x[3]).attach(trainer, 'intra')

    # Add progress bar showing trainer metrics
    mtcs = ['loss', 'acc', 'inter', 'intra']
    ProgressBar(persist=True).attach(trainer, mtcs)
    return trainer


def create_evaluator(model, args):
    cls_loss = torch.nn.BCEWithLogitsLoss()

    def _validation_step(_, batch):
        model.eval()
        with torch.no_grad():
            data = batch_to_tensor(batch, args)

            toks, typs, prefix, tags, events = [x.to(args.device) for x in data]
            mask = torch.not_equal(toks, args.pad_tok_id).to(args.device)

            logit, left_feat = model(toks, typs, prefix, mask)

            loss = cls_loss(logit, tags.float())
            pred = torch.gt(logit, 0.)

            acc = accuracy_score(tags.cpu(), pred.cpu())

            return loss, acc

    evaluator = Engine(_validation_step)
    Average(lambda x: x[0]).attach(evaluator, 'loss')
    Average(lambda x: x[1]).attach(evaluator, 'acc')

    ProgressBar(persist=True).attach(evaluator)
    return evaluator


def create_tester(model, args, msg_feats, ref_num):
    cls_loss = torch.nn.BCEWithLogitsLoss()

    def _test_step(_, batch):
        model.eval()
        with torch.no_grad():
            data = batch_to_tensor(batch, args)

            toks, typs, prefix, tags, events = [x.to(args.device) for x in data]
            mask = torch.not_equal(toks, args.pad_tok_id).to(args.device)

            idx = [a for tag, evt, a, b, fix, tok, _ in batch]
            idx = torch.tensor(idx, dtype=torch.long).to(args.device)

            me = [True if a == b else False for tag, evt, a, b, fix, tok, _ in batch]
            me = torch.tensor(me, dtype=torch.long).to(args.device)

            logit, left_feat = model(toks, typs, prefix, mask)

            loss = cls_loss(logit, tags.float())
            pred = torch.gt(logit, 0.)

            # top_k_values, top_k_indices = torch.topk(torch.sigmoid(logit), k=90, largest=True)#

            msk = torch.gt(torch.sigmoid(logit), args.tao)

            acc = accuracy_score(tags.cpu(), pred.cpu())
            msk = torch.logical_or(msk, me)

            msk_idx, = torch.nonzero(msk, as_tuple=True)
            idx = torch.index_select(idx, dim=0, index=msk_idx)
            # idx = torch.index_select(idx, dim=0, index=top_k_indices)#
            ## feats = (pfx_feat + left_feat) / 2.
            feats = left_feat
            feat = torch.index_select(feats, dim=0, index=msk_idx)
            evt = torch.index_select(events, dim=0, index=msk_idx)

            # feat = torch.index_select(feats, dim=0, index=top_k_indices)#
            # evt = torch.index_select(events, dim=0, index=top_k_indices)#

            msg_feats.index_add_(0, idx.cpu(), feat.cpu())
            ref_num.index_add_(0, idx.cpu(), torch.ones_like(evt, device='cpu'))

            return loss, acc

    tester = Engine(_test_step)

    Average(lambda x: x[0]).attach(tester, 'loss')
    Average(lambda x: x[1]).attach(tester, 'acc')

    ProgressBar(persist=True).attach(tester)
    return tester


def test_on_block(model, cfg, blk, b=0):
    test = blk['test']
    print("Length of test['samples']:", len(test['samples']))

    msg_tags = np.array(test['tw_to_ev'], dtype=np.int32)

    tst_num = msg_tags.shape[0]
    msg_feats = torch.zeros((tst_num, model.feat_size()), device='cpu')  # cfg.feat_dim
    ref_num = torch.zeros((tst_num,), dtype=torch.long, device='cpu')

    train, valid = blk['train'], blk['valid']
    cfg.train_evt_num = len(train['ev_to_idx'])
    # print("cfg.train_evt_num:", cfg.train_evt_num)
    test_gen, test_batch_num = data_generator(test['samples'], cfg.batch_size)
    tester = create_tester(model, cfg, msg_feats, ref_num)

    print("Evaluate model on test data ...")
    test_state = tester.run(test_gen, epoch_length=test_batch_num)

    print("Available metrics:", test_state.metrics.keys())  # Add debug print to check available metrics
    test_metrics = [(m, test_state.metrics[m]) for m in ['loss', 'acc']]
    test_metrics = ", ".join([("%s: %.4f" % (m, v)) for m, v in test_metrics])
    print(f"Test: {test_metrics}\n", flush=True)

    # clustering & report
    msg_feats = msg_feats / (ref_num + torch.eq(ref_num, 0).float()).unsqueeze(-1)
    msg_feats = msg_feats.numpy()
    n_clust = len(test['ev_to_idx'])

    if not os.path.exists(cfg.eva_data):
        os.makedirs(cfg.eva_data)

    Evaluate_datas = {'msg_feats': msg_feats, 'msg_tags': msg_tags, 'n_clust': n_clust}
    if cfg.offline:
        print(f"save Evaluate_datas_offline to '{cfg.eva_data}/evaluate_data_long_tail.npy'", end='\t')
        np.save(f'{cfg.eva_data}/evaluate_data_long_tail.npy', Evaluate_datas)
    else:
        print(f"save Evaluate_datas{b} to '{cfg.eva_data}/evaluate_data_M{b}.npy'", end='\t')
        e_path = cfg.eva_data + f'/evaluate_data_M{b}.npy'
        np.save(e_path, Evaluate_datas)

    print('done')

    k_means_score = run_kmeans(msg_feats, n_clust, msg_tags)
    dbscan_score = run_hdbscan(msg_feats, msg_tags)

    del msg_feats

    return k_means_score, dbscan_score


def load_ckpt(model, args, ckpt, b):
    print(f"Load best ckpt for block {b} from '{ckpt}'")

    ckpt = torch.load(ckpt, map_location=args.device)
    model.load_state_dict(ckpt['model'], strict=False)

    ckpt_args = ckpt['args']
    ckpt_args.dataset = args.dataset
    ckpt_args.plm_path = args.plm_path
    ckpt_args.batch_size = args.batch_size
    ckpt_args.device = args.device
    ckpt_args.tao = args.tao

    return model, ckpt_args


def start_run(cfg, blocks):
    tokenizer = AutoTokenizer.from_pretrained(args.plm_path)
    cfg.pad_tok_id = tokenizer.pad_token_id
    model = get_model(cfg).to(cfg.device)
    # print settings
    print_table([(k, str(v)[0:60]) for k, v in vars(cfg).items()])
    kmeans_scores, dbscan_scores = [], []
    ckpt_list = []

    for b, blk in enumerate(blocks):
        print(f"Processing block-{b}...", flush=True)
        print(f"Block-{b} content keys: {blk.keys()}")

        train, valid, test = (blk[n] for n in ('train', 'valid', 'test'))
        print(
            f"Train samples: {len(train['samples'])}, Valid samples: {len(valid['samples'])}, Test samples: {len(test['samples'])}")

        if b > 0:
            print(f"test model on data block-{b} ...", flush=True)
            kms_score, dbs_score = test_on_block(model, cfg, blk, b)
            kmeans_scores.append(kms_score)
            dbscan_scores.append(dbs_score)

            print("KMeans:")
            print_scores(kmeans_scores)
            print("DBSCAN:")
            print_scores(dbscan_scores)

        if b % 3 == 0:
            gc.collect()
            print(f"train on data block-{b} ...", flush=True)
            model, ckpt = train_on_block(model, cfg, blk, b)
            ckpt_list.append(ckpt)

        if b == 0 and args.offline:
            print(f"close test on data block-{b} ...", flush=True)
            kms_score, dbs_score = test_on_block(model, args, blk, b)
            kmeans_scores.append(kms_score)
            dbscan_scores.append(dbs_score)

            print("KMeans:")
            print_scores(kmeans_scores)
            print("DBSCAN:")
            print_scores(dbscan_scores)

    if args.offline:
        ckpt_list_file = os.path.join(args.ckpt_path, 'best_off_model.txt')
    else:
        ckpt_list_file = os.path.join(args.ckpt_path, 'ckpt_list.txt')

    with open(ckpt_list_file, 'w', encoding='utf8') as f:
        ckpt_list = [str(p) for p in ckpt_list]
        f.write("\n".join(ckpt_list))

    return model


def train_on_block(model, args, blk, blk_id=0):
    # reload plm in tuning mode
    if blk_id > 0 and args.plm_tuning:
        print("accumulate reload PLM parameters", flush=True)
        model.accumulate_reload_plm(args.device)
    train, valid = blk['train'], blk['valid']
    # fewer data for code test
    ###
    # train['samples'] = train['samples'][:500]
    # valid['samples'] = valid['samples'][:200]

    args.train_evt_num = len(train['ev_to_idx'])

    train_gen, train_batch_num = data_generator(train['samples'], args.batch_size, True, True)
    valid_gen, valid_batch_num = data_generator(valid['samples'], args.batch_size, False, True)

    # create model, optimizer and learning rate scheduler
    optimizer, lr_scheduler = initialize(model, args, train_batch_num)

    # print model parameters
    # summary(model, input_size=((args.batch_size, 50), (args.batch_size, 50)))

    # Setup model trainer and evaluator
    trainer = create_trainer(model, optimizer, lr_scheduler, args)
    evaluator = create_evaluator(model, args)

    @trainer.on(Events.EPOCH_STARTED)
    def log_learning_rates(_):
        for param_group in optimizer.param_groups:
            print("{} lr: {:>1.2e}".format(param_group.get('name', ''), param_group["lr"]))

    # Run model evaluation every epoch and show results
    @trainer.on(Events.EPOCH_COMPLETED(every=1))
    def evaluate_model():  # eng
        print("Evaluate model on eval data ...")
        val_state = evaluator.run(valid_gen, epoch_length=valid_batch_num)

        eval_metrics = [(m, val_state.metrics[m]) for m in ['loss', 'acc']]
        eval_metrics = ", ".join([("%s: %.4f" % (m, v)) for m, v in eval_metrics])

        print(f"Eval: {eval_metrics}\n", flush=True)

    def score_function(_):
        if args.early_stop_monitor == 'loss':
            return - evaluator.state.metrics['loss']
        elif args.early_stop_monitor in evaluator.state.metrics:
            return evaluator.state.metrics[args.early_stop_monitor]
        else:
            raise Exception('unsupported metric %s' % args.early_stop_monitor)

    if args.offline:
        filename_prefix = f"{args.model_name}-{'tuning' if args.plm_tuning else 'fixed'}-{args.dataset_name}-offline"
    else:
        filename_prefix = f"{args.model_name}-{'tuning' if args.plm_tuning else 'fixed'}-{args.dataset_name}-{blk_id}"
    ckpt_handler = ModelCheckpoint(args.ckpt_path, score_function=score_function,
                                   filename_prefix=filename_prefix, n_saved=3,
                                   create_dir=True, require_empty=False)

    # if not tuning plm,
    model_state = get_model_state(model, ['pair_cls', 'pfx_embedding'], args.plm_tuning)
    ckpt = {'model': model_state, 'args': CkptWrapper(args)}
    trainer.add_event_handler(Events.EPOCH_COMPLETED(every=1), ckpt_handler, ckpt)

    hdl_early_stop = EarlyStopping(patience=args.early_stop_patience, score_function=score_function, trainer=trainer)
    # Note: the handler is attached to an *Evaluator* (runs one epoch on validation dataset).
    evaluator.add_event_handler(Events.COMPLETED, hdl_early_stop)

    # start training loop
    trainer.run(train_gen, max_epochs=args.max_epochs, epoch_length=train_batch_num)

    # load best checkpoint
    best_ckpt = ckpt_handler.last_checkpoint
    print(f"Load best model checkpoint from '{best_ckpt}'")
    ckpt = torch.load(best_ckpt)
    model.load_state_dict(ckpt['model'], strict=False)
    del ckpt
    return model, best_ckpt


# utils

def load_data_blocks(path_to_data, args, tokenizer):
    print(f"load data from '{path_to_data}'... ", end='')
    dataset = np.load(path_to_data, allow_pickle=True)
    print("\tDone")

    path_to_blocks = []
    print(f"encode block samples, ")

    for i, blk in enumerate(dataset):
        print(f"Message Block{i}", flush=True)
        train, valid, test = (blk[n] for n in ('train', 'valid', 'test'))
        print(
            f"Train samples: {len(train['samples'])}, Valid samples: {len(valid['samples'])}, Test samples: {len(test['samples'])}")

        path = f"{args.file_path}/cache/cache_long_tail/"

        if not os.path.exists(path):
            os.makedirs(path)
        if args.offline:
            # blk_path = os.path.join(path, f"{args.model_name}-{args.dataset_name}-offline.npy")
            blk_path = os.path.join(path, f"{args.model_name}-{args.dataset_name}.npy")
        else:
            blk_path = os.path.join(path, f"{args.model_name}-{args.dataset_name}-M{i + 1}.npy")

        if not os.path.exists(blk_path):
            print("train dateset processing", end=" ")
            train['samples'] = encode_samples(train['samples'], train['data'], tokenizer, args.pmt_feats)
            print("done")

            print("valid dateset processing", end=" ")
            valid['samples'] = encode_samples(valid['samples'], valid['data'], tokenizer, args.pmt_feats)
            print("done")

            print("test dateset processing", end=" ")
            test['samples'] = encode_samples(test['samples'], test['data'], tokenizer, args.pmt_feats)
            print("done")

            torch.save(
                {'train': train, 'valid': valid, 'test': test},
                blk_path
            )
        if blk_path not in path_to_blocks:
            path_to_blocks.append(blk_path)

    del dataset
    print("Done")

    path_to_blocks = ['../model/model_saved/rplmsed/cache/cache_long_tail/roberta-large-twitter12.npy']
    for blk_path in path_to_blocks:
        print(f"load block from '{blk_path}'... \n", end='')
        loaded_blk = torch.load(blk_path)
        # 检查加载的数据是否为字典
        if isinstance(loaded_blk, dict):
            print(f"Loaded block type: {type(loaded_blk)}")
            print(f"Loaded block keys: {loaded_blk.keys()}")
            yield loaded_blk
        else:
            print(f"Error: Loaded block is not a dictionary, but {type(loaded_blk)}")
            yield None


class CkptWrapper:
    def __init__(self, state: Any):
        self.state = state

    def state_dict(self):
        return self.state


def get_model_state(model, params, plm_tuning):
    if plm_tuning:
        return model
    else:
        model_state = model.state_dict()
        state = OrderedDict()
        for k, v in model_state.items():
            for p in params:
                if k.startswith(p):
                    state[k] = model_state[k]
                    break

        return CkptWrapper(state)


def width(text):
    return sum([2 if '\u4E00' <= c <= '\u9FA5' else 1 for c in text])


def print_table(tab):
    col_width = [max(width(x) for x in col) for col in zip(*tab)]
    print("+-" + "-+-".join("{:-^{}}".format('-', col_width[i]) for i, x in enumerate(tab[0])) + "-+")
    for line in tab:
        print("| " + " | ".join("{:{}}".format(x, col_width[i]) for i, x in enumerate(line)) + " |")
    print("+-" + "-+-".join("{:-^{}}".format('-', col_width[i]) for i, x in enumerate(tab[0])) + "-+")


def data_generator(data, batch_size, shuffle=False, repeat=False):
    batch_num = math.ceil(len(data) / batch_size)
    return create_data_generator(data, batch_size, shuffle, repeat, batch_num), batch_num


def create_data_generator(data, batch_size, shuffle, repeat, batch_num):
    while True:
        if shuffle:
            shuffled_idx = [i for i in range(len(data))]
            random.shuffle(shuffled_idx)
            data = [data[i] for i in shuffled_idx]

        batch_id = 0
        while batch_id < batch_num:
            offset = batch_id * batch_size
            batch_data = data[offset:offset + batch_size]
            yield batch_data

            batch_id = batch_id + 1
        if repeat:
            continue
        else:
            break


def pad_seq(seq, max_len, pad=0, pad_left=False):
    """
    padding or truncate sequence to fixed length
    :param seq: input sequence
    :param max_len: max length
    :param pad: padding token id
    :param pad_left: pad on left
    :return: padded sequence
    """
    if max_len < len(seq):
        seq = seq[:max_len]
    elif max_len > len(seq):
        padding = [pad] * (max_len - len(seq))
        if pad_left:
            seq = padding + seq
        else:
            seq = seq + padding
    return seq


def run_kmeans(msg_feats, n_clust, msg_tags):
    # defalut:10
    k_means = KMeans(init="k-means++", n_clusters=n_clust, n_init=40, random_state=0)
    k_means.fit(msg_feats)

    msg_pred = k_means.labels_
    score_funcs = [
        ("NMI", metrics.normalized_mutual_info_score),
        ("AMI", metrics.adjusted_mutual_info_score),
        ("ARI", metrics.adjusted_rand_score),
    ]

    scores = {m: fun(msg_tags, msg_pred) for m, fun in score_funcs}

    return scores


def run_hdbscan(msg_feats, msg_tags):
    hdb = HDBSCAN(min_cluster_size=8)
    hdb.fit(msg_feats)

    msg_pred = hdb.labels_
    score_funcs = [
        ("NMI", metrics.normalized_mutual_info_score),
        ("AMI", metrics.adjusted_mutual_info_score),
        ("ARI", metrics.adjusted_rand_score),
    ]

    scores = {m: fun(msg_tags, msg_pred) for m, fun in score_funcs}

    return scores


def run_dbscan(msg_feats, msg_tags):
    db = OPTICS(min_cluster_size=8, xi=0.01)
    db.fit(msg_feats)

    msg_pred = db.labels_
    score_funcs = [
        ("NMI", metrics.normalized_mutual_info_score),
        ("AMI", metrics.adjusted_mutual_info_score),
        ("ARI", metrics.adjusted_rand_score),
    ]

    scores = {m: fun(msg_tags, msg_pred) for m, fun in score_funcs}

    return scores


def print_scores(scores):
    line = [' ' * 4] + [f'   M{i:02d} ' for i in range(1, len(scores) + 1)]
    print("".join(line))

    score_names = ['NMI', 'AMI', 'ARI']
    for n in score_names:
        line = [f'{n} '] + [f'  {s[n]:1.3f}' for s in scores]
        print("".join(line))
    print('\n', flush=True)


def encode_samples(samples, raw_data, tokenizer, pmt_idx):
    data = []
    for tag, ev_idx, (tw_a, tw_b), pmt_feat in samples:
        tw_a_text = raw_data[tw_a].text
        tw_b_text = raw_data[tw_b].text
        tok = tokenizer(tw_a_text, tw_b_text, padding=True)

        # 只保留需要的关联特征
        # (entities, hashtags, user, words, time)
        pmt_feat = [pmt_feat[f] for f in pmt_idx]

        base = [2 * i for i in range(len(pmt_feat))]
        pmt_ids = [b + f for f, b in zip(pmt_feat, base)]

        if 'token_type_ids' not in tok:
            types = [0, 0, 1, 1]
            token_type_ids = tok.encodings[0].sequence_ids
            j = 0
            for i, t in enumerate(token_type_ids):
                if t is None:
                    token_type_ids[i] = types[j]
                    j += 1
        else:
            token_type_ids = tok['token_type_ids']

        data.append((tag, ev_idx, tw_a, tw_b, pmt_ids, tok['input_ids'], token_type_ids))

    return data


def count_condition(data, key, threshold):
    return sum(entry[key] > threshold for entry in data), sum(entry[key] <= threshold for entry in data)


def calculate_average_min_score(newscore, min_score, max_score):
    for i, score in enumerate(newscore):
        for key, value in score.items():
            min_score[i][key] = min(min_score[i][key], value)
            max_score[i][key] = max(max_score[i][key], value)

    return min_score, max_score


class StructAttention(torch.nn.Module):
    """
    The class is an implementation of the paper A Structured Self-Attentive Sentence Embedding
    """

    def __init__(self, feat_dim, hid_dim, att_head_num=1):
        """
        Initializes parameters suggested in paper
        Args:
            feat_dim:       {int} hidden dimension for lstm
            hid_dim:        {int} hidden dimension for the dense layer
            att_head_num:   {int} attention-hops or attention heads
        Returns:
            self
        Raises:
            Exception
        """
        super(StructAttention, self).__init__()
        self.W1 = torch.nn.Linear(feat_dim, hid_dim, bias=False)
        nn.init.xavier_normal_(self.W1.weight)

        self.W2 = torch.nn.Linear(hid_dim, att_head_num, bias=False)
        nn.init.xavier_normal_(self.W2.weight)

        self.att_head_num = att_head_num

    def forward(self, inpt, mask=None):
        """
        :param inpt: [len, bsz, dim]
        :param mask: [len, bsz]
        :return: [bsz, head_num, dim], [bsz, head_num, len]
        """
        hid = torch.tanh(self.W1(inpt))
        hid = self.W2(hid)

        if mask is not None:
            mask = mask.float().unsqueeze(-1).expand(-1, -1, self.att_head_num)
            mask = (1. - mask) * 1e10
            hid = hid - mask
        att = torch.softmax(hid, dim=0).permute(1, 2, 0)

        outp = att @ inpt.permute(1, 0, 2)

        return outp, att


class PairPfxTuningEncoder(nn.Module):
    def __init__(self, pmt_len,
                 plm_path, plm_tuning=False, from_config=False,
                 use_ctx_att=True, ctx_att_head_num=2):
        super().__init__()
        self.pfx_len = pmt_len
        self.plm_path = plm_path

        if from_config:
            config = AutoConfig.from_pretrained(plm_path)
            self.plm = AutoModel.from_config(config)
        else:
            self.plm = AutoModel.from_pretrained(plm_path)

        if not plm_tuning:
            for name, param in self.plm.named_parameters():
                param.requires_grad = False
                param.detach_()

        self.plm_oupt_dim = self.plm.config.hidden_size

        self.plm_emb_dim = self.plm.embeddings.word_embeddings.embedding_dim

        self.pfx_embedding = nn.Embedding(self.pfx_len * 2, self.plm_emb_dim)
        self.pfx_mask = torch.ones((1, self.pfx_len), dtype=torch.bool)

        self.linear = nn.Linear(self.plm_oupt_dim, self.plm_oupt_dim // 2)

        self.ctx_att = None
        if use_ctx_att:
            self.ctx_att = StructAttention(self.plm_oupt_dim // 2, self.plm_oupt_dim // 4,
                                           att_head_num=ctx_att_head_num)
        self.pair_cls = nn.Linear(2 * (self.plm_oupt_dim // 2), 1)

    def feat_size(self):
        return self.plm_oupt_dim // 2

    def reload_plm(self, device):
        self.plm = AutoModel.from_pretrained(self.plm_path).to(device)

    # 0.4
    def accumulate_reload_plm(self, device, accumulate_rate=0.4):
        origin = AutoModel.from_pretrained(self.plm_path).to('cpu')
        plm_params = self.plm.named_parameters()
        origin_params = origin.named_parameters()
        for ((tgt_name, tgt_param), (src_name, src_param)) in zip(plm_params, origin_params):
            assert (tgt_name == src_name), f"param name {tgt_name} and {src_name} does not match"
            tgt_param.data = (1. - accumulate_rate) * tgt_param.data + accumulate_rate * src_param.to(device).data

    def fix_plm(self):
        for name, param in self.plm.named_parameters():
            param.requires_grad = False
            param.detach_()

    def forward(self, inputs, types, prompt, mask):
        bsz, txt_len = mask.size()

        pmt_msk = self.pfx_mask.to(inputs.device).expand(bsz, -1)
        ext_msk = torch.cat([pmt_msk, mask], dim=-1)
        # ext_msk =mask#

        pmt_emb = self.pfx_embedding(prompt)
        pmt_len = prompt.size(-1)
        txt_emb = self.plm.embeddings(inputs)
        embed = torch.cat([pmt_emb, txt_emb], dim=1)
        # embed= txt_emb #
        att_msk = ext_msk[:, None, None, :]
        att_msk = (1.0 - att_msk.float()) * torch.finfo(torch.float).min
        plm_oupt = self.plm.encoder(embed, att_msk, output_hidden_states=True)

        hidden = plm_oupt['last_hidden_state']
        # if self.ctx_att is not None:
        hidden = torch.tanh(self.linear(hidden))

        pmt_feat = hidden[:, :pmt_len, ...]
        tok_feat = hidden[:, pmt_len:, ...]
        # tok_feat = hidden#

        left_msk = (1 - types) * mask
        left_feat = tok_feat * left_msk.unsqueeze(-1)
        left_msk = torch.cat([pmt_msk.int(), left_msk], dim=1)
        left_feat = torch.cat([pmt_feat, left_feat], dim=1)
        if self.ctx_att is None:
            left_feat = left_feat.sum(dim=-2) / left_msk.sum(-1, keepdims=True)
        else:
            left_feat, left_att = self.ctx_att(left_feat.permute(1, 0, 2), mask=left_msk.permute(1, 0))
            left_feat = torch.mean(left_feat, dim=1)

        right_msk = types * mask
        right_feat = tok_feat * right_msk.unsqueeze(-1)
        if self.ctx_att is None:
            right_feat = right_feat.sum(dim=-2) / right_msk.sum(-1, keepdims=True)
        else:
            right_feat, right_att = self.ctx_att(right_feat.permute(1, 0, 2), mask=right_msk.permute(1, 0))
            right_feat = torch.mean(right_feat, dim=1)

        cls_feat = torch.cat([left_feat, right_feat], dim=-1)

        logit = self.pair_cls(cls_feat).squeeze(dim=-1)

        return logit, left_feat


if __name__ == '__main__':
    args = args_define().args
    dataset = DatasetLoader("maven").load_data()
    rplmsed = RPLM_SED(args, dataset)

    # rplmsed.preprocess()
    rplmsed.fit()
    predictions, ground_truths = rplmsed.detection()

    # Evaluate model
    rplmsed.evaluate(predictions, ground_truths)
